/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

#include "common.h"
#include "criticalbands.h"

//
//  These functions implement the critical band analysis used by Hermansky in the
//  calculations of PLP. These functions ONLY implement the frequency warping.
//  Hynek Hermansky, Perceptual Linear Predictive (PLP) Analysis of speech, JASA vol 87, April 90, pp 1738-1752
//
                                                                                                                          
double bark2hertz(double fBark)
     // Convertit des Barks en Hertz
{
  return(600.*sinh((float)(fBark/6.)));
}

double hertz2bark(double hertz)
     // Convertit des Hertz en Barks
{
  return(6.* log((hertz/(600.))+sqrt(pow((hertz/(600.)),2.)+1.)));
}

double hertz2mel(double hertz)
{
  return(log(1.+ hertz/700.)*1127.01048); // ln(1 + f/700) * 1000 / ln(1+1000/700)
}

double mel2hertz(double mel)
{
  return(700.*(exp(mel/1127.01048)-1.));
}


static double a = 3000. * (1. - 3000. / 8000.) - log(3000.) * 500.,
              b = 3000. * (1. - 3000. / 8000.);

double hertz2mely(double hertz)
{
  if (hertz < 3000) return(hertz * (1. - hertz / 8000.));
  else return  log(hertz) * 500. + a;
}

double mely2hertz(double mel)
{
  if (mel < b) return(8000./2. * (1. - sqrt(1. - 4. * mel / 8000.)));
  else return(exp((mel - a) / 500.));
}

// Critical band as defined by Hermansky (the symetry due to convolution is involved
static float criticalBand(double w)
{
  if (w < -2.5) return 0.;
  else if (w < -0.5) return exp(log(10.) * (0.5 + w));
  else if  (w < 0.5) return 1.;
  else if (w < 1.3) return exp(log(10.) * 2.5 * (0.5 - w));
  else return 0.;
}


bool checkMelFilterWidth(
 FrequencyScale frequencyScale,
 int nbBands,
 int fftOrder,
 double samplingRate)
{
  double highestPerceptualFrequency,
         widthHz;
  double (*hertz2percep)(double),(*percep2hertz)(double);

  switch (frequencyScale) {
  case Bark:
    hertz2percep = hertz2bark;
    percep2hertz = bark2hertz;
    break;
  case Mel:
    hertz2percep = hertz2mel;
    percep2hertz = mel2hertz;
    break;
  case Mely:
    hertz2percep = hertz2mely;
    percep2hertz = mely2hertz;
    break;
  }
  highestPerceptualFrequency = hertz2percep(samplingRate / 2.);
  widthHz = percep2hertz(highestPerceptualFrequency / nbBands);

  return (widthHz / (samplingRate / fftOrder) > 1.); // false if the width is too small
}

void buildOneMelFilter(
  float freqCenterHertz, // frequency of the filter center in hertz
  float percepWidth,     // width expressed in the perceptive scale
  float filter[],        // the resulting filter
  int sizeOfSpectrum,    // Number of points in the spectrum to be filtered.
  double samplingRate,
  double (*hertz2percep)(double),
  double (*percep2hertz)(double))
{
  int l, cleft, cright, center;
  double sum;
  float filterFreqMel = hertz2percep(freqCenterHertz),
        widthLeftHz, widthRightHz;

  widthLeftHz = freqCenterHertz  - percep2hertz(filterFreqMel - percepWidth);
  widthRightHz = -freqCenterHertz  + percep2hertz(filterFreqMel + percepWidth);
  cleft = IROUND((*percep2hertz)(filterFreqMel - widthLeftHz) / samplingRate * sizeOfSpectrum);
  cleft = max(0, cleft);
  cright = IROUND(percep2hertz(filterFreqMel + widthRightHz) / samplingRate * sizeOfSpectrum);
  cright = min(sizeOfSpectrum / 2, cright);
  center = IROUND(freqCenterHertz / samplingRate * sizeOfSpectrum);
  for (l = cleft, sum = 0.; l < cright; l++) {
    if (l < center) {
      float dhertz = freqCenterHertz -  l * samplingRate / sizeOfSpectrum;
      filter[l] = max(1 - dhertz / widthLeftHz, 0.);
    }
    else {
      float dhertz = l * samplingRate / sizeOfSpectrum - freqCenterHertz;
      filter[l] = max(1 - dhertz / widthRightHz, 0.);
    }
    sum += filter[l];
  }

  // Normalization
  if (fabs(sum) > EPSILON) for (l = cleft; l < cright; l++) filter[l] /= sum;
}

void buildOneBarkFilter(
  float freqCenterHertz, // frequency of the filter center in hertz
  float filter[],
  int sizeOfSpectrum,    // Number of points in the spectrum to be filtered.
  double samplingRate,
  double (*hertz2percep)(double),
  double (*percep2hertz)(double))
{
  int l, cleft, cright;
  double sum;
  float filterFreqBk = hertz2percep(freqCenterHertz);

  cleft = IROUND(percep2hertz(filterFreqBk - 2.5) / samplingRate * sizeOfSpectrum);
  cleft = max(0, cleft);
  cright = IROUND(percep2hertz(filterFreqBk + 1.3) / samplingRate * sizeOfSpectrum);
  cright = min(sizeOfSpectrum / 2, cright);

  for (l = cleft, sum = 0.; l < cright; l++) {
    float b = hertz2percep(l * samplingRate / sizeOfSpectrum) - filterFreqBk;
    filter[l] = criticalBand(b);
    sum += filter[l];
  }
  // normalization
  if (fabs(sum) > EPSILON) for (l = cleft; l < cright; l++) filter[l] /= sum;
}


float **buildCriticalBandFilters(
  double samplingRate,
  int nbFilters,
  int FFTorder,
  int **centerFrequencies,
  bool ASRstrategy,         // if true the filters frequency are linear in bark (like ASR)
  FrequencyScale scale,      // perceptive frequency scale
  MelStrategy melStrategy,
  bool nonLinearDisplayFrequency)    // if true the display frequency is bark or mel.
{
  float **filters  = new float* [nbFilters];
  int i;
  float highestPerceptualFrequency, perceptiveWidth;
  double (*hertz2percep)(double),(*percep2hertz)(double);

   switch (scale) {
     case Bark:
       hertz2percep = hertz2bark;
       percep2hertz = bark2hertz;
       break;
     case Mel:
       hertz2percep = hertz2mel;
       percep2hertz = mel2hertz;
       break;
     case Mely:
       hertz2percep = hertz2mely;
       percep2hertz = mely2hertz;
       break;
   }

   highestPerceptualFrequency = hertz2percep(samplingRate / 2.);

   switch (melStrategy) {
     case VariableMelWidth:
       perceptiveWidth =  highestPerceptualFrequency / (nbFilters + 1.);
       break;
     case ConstantMelWidth:
       perceptiveWidth = 100.;
       break;
   }

  *centerFrequencies = new int[nbFilters];
  for (i = 0; i < nbFilters; i++) {
    double filterFreqHz, sum;
    int cleft, cright;      // left and right coefficient indices of the FFT to be taken into account
    int l;
    filters[i] = new float[FFTorder / 2];
    for (int k = 0; k < FFTorder / 2; k++) filters[i][k]=0.;

    if (ASRstrategy || nonLinearDisplayFrequency)
       filterFreqHz = percep2hertz(i * highestPerceptualFrequency / (nbFilters - 1.));
    else filterFreqHz = i * samplingRate / 2. / (nbFilters - 1.);

    if (nonLinearDisplayFrequency)
       (*centerFrequencies)[i] = IROUND(((float) i) / (nbFilters - 1.) * FFTorder / 2.);
    else (*centerFrequencies)[i] = IROUND(filterFreqHz / samplingRate * FFTorder);

    switch (scale) {
     case Mel: case Mely:
       buildOneMelFilter(filterFreqHz, perceptiveWidth, filters[i], FFTorder, samplingRate, hertz2percep, percep2hertz);
       break;
     case Bark:
       buildOneBarkFilter(filterFreqHz, filters[i], FFTorder, samplingRate, hertz2percep, percep2hertz);
       break;
     }
  }
  return filters;
}


void freeCriticalBandFilters(
  float **filters,
  int nbFilters)
{
   for(int i = 0; i < nbFilters; i++) delete [] filters[i];
   delete [] filters;
}

// rescaling of the spectrum (from the nonlinear scale to the linear scale)
void rescalePerceptiveSpectrum(
  double perceptiveSpectrum[],		// spectrum in the perceptive scale
  int spectrumSize,
  bool melFrequencyScale,	// Mel scale if true
  double samplingRate,
  float rescaledSpectrum[],
  int rescaleSpectrumSize)
{
  double highestPerceptualFrequency,
         lastx;

  if (melFrequencyScale) highestPerceptualFrequency = hertz2mel(samplingRate / 2.);
  else highestPerceptualFrequency = hertz2bark(samplingRate / 2.);

  for (int l = 0; l < rescaleSpectrumSize; l++) rescaledSpectrum[l] = -INFINI_F / 2;
  // first pass to get values defined. If there are more than one spectral value the max is kept in
  // order to get sharp peaks.
  for(int i = 0; i < spectrumSize; i++) {
    double filterFreqHz;
    int k;
    if (melFrequencyScale) filterFreqHz = mel2hertz(((float) i) / spectrumSize * highestPerceptualFrequency);
    else filterFreqHz = bark2hertz(((float) i) / spectrumSize * highestPerceptualFrequency);
    k = min(IROUND(filterFreqHz / samplingRate * rescaleSpectrumSize * 2.), rescaleSpectrumSize - 1);
    rescaledSpectrum[k] = max(perceptiveSpectrum[i], rescaledSpectrum[k]);
  }

  // the second pass is the spectrum completion.
  for(int i = 0; i < rescaleSpectrumSize;) {
    int k;
    // search for a relevant value
    for(k = i; rescaledSpectrum[k] <= -INFINI_F / 3 && k < spectrumSize; k++);
    if (!i) { // first value
      for (; i <= k; i++) rescaledSpectrum[i] = rescaledSpectrum[k];
      i = k + 1;
      lastx = (double) k;
    }
    else if (k == spectrumSize) {     // no more spectral value available
      double lasty = rescaledSpectrum[(int) lastx];
      for (; i < rescaleSpectrumSize; i++) rescaledSpectrum[i] = lasty;
    }
    else if (k == i) {
      lastx = (double) i;
      rescaledSpectrum[i++] = rescaledSpectrum[k];
    }
    else {
      double lasty = rescaledSpectrum[(int) lastx],
             nexty = rescaledSpectrum[k];
      for (; i <= k; i++)
        rescaledSpectrum[i] = lasty + (nexty - lasty) / (k - lastx) * (i - lastx);
      i = k + 1;
      lastx = (double) k;
    }
  }
}

void applyCriticalBands(
  FLOATbuf& X,
  FLOATbuf& Y,
  int FFTorder,
  int cmax,
  float **filters,
  int indices[],
  int nbFilters,
  bool ASRstrategy)
{
  int i, k;
  bool tobecompleted = (nbFilters < FFTorder / 2) || ASRstrategy;
  // tobecompleted true means that there are fewer spectral points calculated than the # of  original spectral points.
  // The algorithms thus interpolates unknow values. In the case of "ASR like", i.e. Automatic Speech Rcognition there
  // is no interpolation, actually only a simple copy of known values, to keep the stepwise calculation.
  float lastY;

  if (tobecompleted) k = 0;

  for (i = 0; i < nbFilters && ((i < nbFilters - 1) ? (indices[i] + indices[i + 1]) / 2 <= cmax : true); i++) {
    int j;
    float x;
    for (j = 0, x = 0.; j < FFTorder / 2; j++)
      x += X[j] * filters[i][j];
    if (tobecompleted) {
      if (ASRstrategy) // no interpolation to show the spectrum as it is used by ASR
        for (; k < ((i < nbFilters - 1) ? ((indices[i] + indices[i + 1]) / 2) : (FFTorder / 2)); k++) Y[k] = x;
      else {
        if (i) for (; k < indices[i]; k++) Y[k] = lastY + (k - indices[i - 1]) / ((float)(indices[i] - indices[i - 1])) * (x - lastY);
        lastY = x;
      }
    }
    else Y[i] = x;
  }
  if (tobecompleted && k < cmax) {
    double y = Y[k - 1];
    for(; k < cmax; k++) Y[k] = y;
  }
}

void spectralCompletion(
  float *Y,   // the result after the computation of critical bands and then the complete spectrum (with FFTorder / 2 points).
  int FFTorder,
  int cmax,
  int indices[],
  int nbFilters,
  bool ASRstrategy)
{
  float *Z = new float [FFTorder/2];
  float lastY;
  int k = 0;

  // copy the original incomplete spectrum
  for (int i = 0; i < nbFilters; i++) Z[i] = Y[i];

  for (int i = 0; i < nbFilters && ((i < nbFilters - 1) ? (indices[i] + indices[i + 1]) / 2 <= cmax : true); i++) {
    float x = Z[i];
    if (ASRstrategy) // no interpolation to show the spectrum as it is used by ASR
      for (; k < ((i < nbFilters - 1) ? ((indices[i] + indices[i + 1]) / 2) : (FFTorder / 2)); k++) Y[k] = x;
    else {
      if (i) for (; k < indices[i]; k++)
      	Y[k] = lastY + (k - indices[i - 1]) / ((float)(indices[i] - indices[i - 1])) * (x - lastY);
      lastY = x;
    }
  }
  if (k < cmax) {
    float y = Y[k - 1];
    for(; k < cmax; k++) Y[k] = y;
  }
  delete [] Z;
}

void applyCriticalBandsff(
  float *X,
  float *Y,
  int FFTorder,
  int cmax,
  float **filters,
  int indices[],
  int nbFilters)
{
  int i, k;

  for (i = 0; i < nbFilters && ((i < nbFilters - 1) ? (indices[i] + indices[i + 1]) / 2 <= cmax : true); i++) {
    int j;
    float x;
    for (j = 0, Y[i] = 0.; j < FFTorder / 2; j++)
      Y[i] += X[j] * filters[i][j];
  }
}

void OLDapplyCriticalBandsff(
  float *X,
  float *Y,
  int FFTorder,
  int cmax,
  float **filters,
  int indices[],
  int nbFilters,
  bool ASRstrategy)
{
  int i, k;
  bool tobecompleted = (nbFilters < FFTorder / 2) || ASRstrategy;
  float lastY;

  if (tobecompleted) k = 0;

  for (i = 0; i < nbFilters && ((i < nbFilters - 1) ? (indices[i] + indices[i + 1]) / 2 <= cmax : true); i++) {
    int j;
    float x;
    for (j = 0, x = 0.; j < FFTorder / 2; j++)
      x += X[j] * filters[i][j];
    if (tobecompleted) {
      if (ASRstrategy) // no interpolation to show the spectrum as it is used by ASR
        for (; k < ((i < nbFilters - 1) ? ((indices[i] + indices[i + 1]) / 2) : (FFTorder / 2)); k++) Y[k] = x;
      else {
        if (i) for (; k < indices[i]; k++) Y[k] = lastY + (k - indices[i - 1]) / ((float)(indices[i] - indices[i - 1])) * (x - lastY);
        lastY = x;
      }
    }
    else Y[i] = x;
  }
  if (tobecompleted && k < cmax) {
    int y = Y[k - 1];
    for(; k < cmax; k++) Y[k] = y;
  }
}

// Dct and Idct transforms (see http://en.wikipedia.org/wiki/Discrete_cosine_transform for instance)
float *dct(
  float *x,
  float *y,
  int n)
{
  for (int j = 0; j < n; j++) {
    y[j] = 0.;
    for (int k = 0; k < n; k++)
       y[j] += x[k] * cos(_PI / n * j * (k + 1. / 2.));
  }
  return (y);
}

#ifdef UIY
float *idct(
  float *x,
  float *y,
  int n,
  int nlifter,
  int nbSamplesDisp,
  bool completionStrategy, // true if ASRlike, i.e. not linear;
  bool frequencyScale, // true if Mel, false if Bark
  double samplingRate)
{
  for(int j = 0; j < n; j++) {
    y[j] = x[0] / 2.;

   for (int k = 1; k < n; k++)
      y[j] += x[k] * cos(_PI / n * (j + 1. / 2.) * k);
    y[j] *= 2. / n;
  }
  return(y);
}
#endif

float *idct(
  float *x,
  float *y,
  int n,
  int nlifter,
  int nbSamplesDisp,
  bool completionStrategy, // true if ASRlike, i.e. not linear;
  double samplingRate,
  double (*hertz2percep)(double))
{
  for(int j = 0; j < nbSamplesDisp; j++) {
    float omega;
    if (completionStrategy)
      omega = hertz2percep(j * samplingRate / 2. / nbSamplesDisp) / hertz2percep(samplingRate / 2.);
    else omega = j / ((float) nbSamplesDisp);
    y[j] = x[0] / 2.;

    for (int k = 1; k < nlifter; k++)
      y[j] += x[k] * cos(_PI / n * (omega * n + 1. / 2.) * k);
    y[j] *= 2. / n;
  }
  return(y);
}


#ifdef SCILAB
function [y]= criticalband (w)
s=size(w)
for i=1:s(2) if w(i) < -1.3  then
  y(1,i) = 0
elseif w(i) < -0.5  then
  y(1,i) = exp(log(10) * 2.5 .* (w(i) + 0.5))
elseif  w(i) < 0.5 then
  y(1,i) = 1
elseif w(i) < 2.5
  y(1,i) = exp(log(10) * (0.5 - w(i)))
else
  y(1,i) = 0
end
end

MFCCi= somme de k=0  20 Xk * cos(i*(k-1/2)*Pi/20)

// Dct and Idct transform (see http://en.wikipedia.org/wiki/Discrete_cosine_transform for instance)
function [y] = dct (x)
s = size(x)
for j=0:s(2)-1
  y(j+1) = 0.
  for k=0:s(2)-1
    y(j+1) = y(j+1) + x(k+1) * cos(%pi / s(2) * j * (k + 1/2))
//    disp(y(j+1))
  end
end
endfunction

function [y] = idct(x)
s = size(x)
for j=0:s(2)-1
  y(j+1) = x(1) / 2
  for k=1:s(2)-1
    y(j+1) = y(j+1) + x(k+1) * cos(%pi / s(2) * (j + 1 / 2) * k)
 end
  y(j+1) = y(j+1) * 2 / s(2);
 end
endfunction
#endif

