/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

// fichier edit.cpp
//
//
// Primitives d'edition d'un fichier de parole
//

#include "common.h"
#pragma hdrstop

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <io.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "editer.h"
#include "audio.h"
#include "notetiq.h"
#include "outils.h"
#include "files.h"
#include "signal.h"
#include "polypoint.h"
#include "guiklat.h"  // pour reset formants
#include "filterOLA.h"
#include "spectro.h"
#include "pitchm.h"    // pour reset pitch

#define PASTEBUF_LEN 20000
static SAMPLEbuf PasteBuf_(1,"PasteBuf in edit");

/////////////////////////////////////////////////
// quelques remarques sur le signal en cours d'dition
//
// - on peut avoir ct.d_e!=0 (cas des grands fichiers)
// - en non ZOOM, on peut avoir ct.d_a!=ct.d_e
// - le signal en cours est constitu
//     du fichier avant ct.d_e
//     du buffer de ct.d_e  ct.f_e
//     du fichier de infoEdit.derEchtFic  ct.tf-1
// - infoEdit.premEchtFic est inutilis, car par principe
//     il doit toujours tre gal  ct.d_e
// - la valeur max pour ct.f_e est
//       ct.f_e + ct.tf - infoEdit.derEchtFic -1
// - si yaDesModifs==FAUX alors infoEdit.derEchtFic=ct.f_e
// - la taille max de l'cran en chantillon est
//       MAXECHTECR
//         qui dpend de en_cours.frequence
// - la taille max du buffer space est
//       MAXECH
//         qui dpend de en_cours.frequence
//

/////////////////////////////////////////////////
// note sur le presse-papier de winsnoori
//
// le presse-papier est le fichier cutfile.wav
//     dans le directory tempdir
// le fichier cutfile_.wav a une fonction similaire
//     mais est cach de l'utilisateur. Il est utilis
//     pour sauver des portions de parole que la
//     commande "annuler" pourrait rcuprer, mais qui
//     ne sont pas destins au presse-papier.
// infoEdit.cut est la structure infoFile utilise par
//     cutfile.wav et cutfile_.wav, commune au deux
//     sauf pour infoEdit.cut.nom qu'il faut mettre  
//     chaque criture dans l'un ou l'autre.

/////////////////////////////////////////////////
// notes diverses
//
// - infoEdit est utilis soit en direct, soit
//     en indirect transmis en paramtres. Dans tous les
//     cas, il s'agit du mme objet. Il n'y a qu'un seul
//     infoEdit dans tout WinSnoori.

/////////////////////////////////////////////////
// procdures de transferts buffer/fichier
//   concernent des fichiers dj ouverts
//     avec l' ordre hmmioOpen()
//   et une portion de buffer mmoire de type SAMPLEbuf
//
// copyHFileToBuf()    de fichier  buffer
// copyBufToHFile()    de buffer  fichier
// copyHFileToHFile()  de fichier  fichier
//
// la taille  transfrer est donne par
//     bcount : nombre d'octets  transfrer
//
// le buffer est dsign par
//     une variable de type SAMPLEbuf
//     et un dplacement xxx_debut (en SAMPLEs)
//
// chaque fichier est dsign par 
//     une structure de type infoFile
// et utilise les informations suivantes
//     nom : le nom du fichier
//     nb_bits : nombre de bits de chaque SAMPLE
//     swap : ==SWAPINFILE si les octets SAMPLE sont swaps
//
// le fichier source est  positionner avec l'ordre
//       mmioSeek(srcinfo.hmmio,srcoffset,SEEK_SET)
//
// le fichier destination est dj positionn
//
static long copyHFileToBuf(
		 infoFile& srcinfo
                   ,long srcoffset // en octets
		,SAMPLEbuf& bufdst
		   ,long dst_debut // en SAMPLEs
		,long bcount) // en octets
// en retour, le nombre d'octets lus
//     si ==-1 alors erreur
{ int n;
  long i,nbread,nsread;
  char proc[]="copyHFileToBuf()";

  if (debug==2)
  { char ch[80];
    Wmsg->wr(proc);
    sprintf(ch,"... srcinfo: %s, offset=%ld"
              ,srcinfo.nom,srcoffset);
    Wmsg->wr(ch);
    sprintf(ch,"...... swap=%d, nb_bits%d"
	      ,srcinfo.swap,srcinfo.nb_bits);
    Wmsg->wr(ch); 
    sprintf(ch,"... buffer offset=%ld"
	      ,dst_debut);
    Wmsg->wr(ch);
    sprintf(ch,"... bcount=%ld"
	      ,bcount);
    Wmsg->wr(ch);
  }
  if ( (bcount>(bufdst.size()-dst_debut)
	       *sizeof(SAMPLE)) )
  { erboxmmio(1,proc,srcinfo.nom
	     ,"bug in copyHFileToBuf()\ntry to put data after end of buffer");
    return -1;
  }
  
  if (erboxmmioSeek(2,proc,srcinfo.nom
		,mmioSeek(srcinfo.hmmio,srcoffset,SEEK_SET))
	      !=0) return -1;
  erboxmmioRead(3,proc,srcinfo.nom
		    ,nbread=mmioRead(srcinfo.hmmio
		       ,(HPSTR)&bufdst[dst_debut]
		       ,bcount));
  if (nbread==-1) return -1; // scurit

  nsread=nbread/sizeof(SAMPLE);   
  if (srcinfo.swap!=SWAPINFILE)
    swap((char *) &bufdst[dst_debut] ,nbread );
  if ((n=NB_BITS-srcinfo.nb_bits)>0) // left shift
  { for (i = 0; i < nsread; i++) 
      bufdst[dst_debut+i] = bufdst[dst_debut+i] << n;
  }
  return nbread;
}
long copyBufToHFile(
		 SAMPLEbuf& bufsrc
		   ,long src_debut // en SAMPLEs
		,infoFile& dstinfo
		,long bcount) // en octets
// en retour, nombre d'octets transfrs
//     erreur si ==-1
{ int i,diff,nshift;
  long nb,nbwrite=0,scount;
  char proc[]="copyBufToHFile()";

  if (debug==2)
  { char ch[80];
    Wmsg->wr(proc);
    sprintf(ch,"... buffer offset=%ld"
              ,src_debut);
    Wmsg->wr(ch);
    sprintf(ch,"... dstinfo: %s"
	      ,dstinfo.nom);
    Wmsg->wr(ch);
    sprintf(ch,"...... swap=%d, nb_bits%d"
	      ,dstinfo.swap,dstinfo.nb_bits);
    Wmsg->wr(ch); 
    sprintf(ch,"... bcount=%ld"
	      ,bcount);
    Wmsg->wr(ch);
  }
  scount=bcount/sizeof(SAMPLE);
  nshift=NB_BITS-dstinfo.nb_bits;
  if ((nshift==0)&&(dstinfo.swap==SWAPINFILE))
  { // ni shift, ni swap
    // on recopie directement 
    erboxmmioWrite(1,proc,dstinfo.nom
		    ,nbwrite=mmioWrite(dstinfo.hmmio
		       ,(HPSTR)&bufsrc[src_debut]
		       ,bcount));
  }
  else
  { // shift ou swap
    // on passe par le buffer PasteBuf
    long done = src_debut;
    SAMPLE *PasteBuf;

    PasteBuf_.newsize(PASTEBUF_LEN);
    PasteBuf=(SAMPLE *)&PasteBuf_[0];
    while ((scount!=0)&&(nbwrite!=-1))
    { diff=(int)min(PasteBuf_.size(),scount);
      scount-=diff;

      // right shift
      { for (i = 0; i < diff; i++) 
          PasteBuf_[i] = bufsrc[done + i] >> nshift;
      }
      if (dstinfo.swap)
        swap((char *) &PasteBuf[0]
             ,diff * sizeof(SAMPLE) );
      if (erboxmmioWrite(1,"copyBufToHFile()",dstinfo.nom
		    ,nb=mmioWrite(dstinfo.hmmio
			  ,(HPSTR)&PasteBuf[0]
			  ,diff * sizeof(SAMPLE)))
		    !=0) nbwrite=-1;
                    else nbwrite+=nb;
      done += diff;
    }
    PasteBuf_.newsize(1);
  }
  return nbwrite;
}
static long copyHFileToHFile(
	     infoFile& srcinfo
	       ,long srcoffset // en octets
	    ,infoFile& dstinfo
	    ,long bcount) // en octets
// copie bcount octets du fichier src dans le fichier dst
// les deux fichiers sont dj ouverts
// en retour nombre d'octets transfrs
//     erreur si -1
//  initialiser dans srcinfo et dstinfo
//       hmmio, nb_bits, swap
{ long nbread=0,nb,ns;
  int i;
  int nshift;
  char proc[]="copyHFileToHFile()";
  SAMPLE *PasteBuf;

  if (debug==2)
  { char ch[80];
    Wmsg->wr(proc);
    sprintf(ch,"... srcinfo: %s, offset=%ld"
              ,srcinfo.nom,srcoffset);
    Wmsg->wr(ch);
    sprintf(ch,"...... swap=%d, nb_bits%d"
	      ,srcinfo.swap,srcinfo.nb_bits);
    Wmsg->wr(ch); 
    sprintf(ch,"... dstinfo: %s"
	      ,dstinfo.nom);
    Wmsg->wr(ch);
    sprintf(ch,"...... swap=%d, nb_bits%d"
	      ,dstinfo.swap,dstinfo.nb_bits);
    Wmsg->wr(ch); 
    sprintf(ch,"... bcount=%ld"
	      ,bcount);
    Wmsg->wr(ch);
  }
  if (erboxmmioSeek(13,proc,srcinfo.nom
		,mmioSeek(srcinfo.hmmio,srcoffset,SEEK_SET))
	      !=0) nbread=-1;
  PasteBuf_.newsize(PASTEBUF_LEN);
  PasteBuf=(SAMPLE *)&PasteBuf_[0];
  while ((bcount!=0)&&(nbread!=-1))
  { if (erboxmmioRead(14,proc,""
		 ,nb=mmioRead(srcinfo.hmmio
		    ,(HPSTR)&PasteBuf[0]
		    ,min(bcount,PasteBuf_.size()
			   *sizeof(SAMPLE)) ))
	        !=0) nbread=-1;
    else
    { // swap et shift de la source
      if (srcinfo.swap!=SWAPINFILE)
        swap((char *) &PasteBuf[0] ,nb );
      ns=nb/sizeof(SAMPLE);
      if ((nshift=NB_BITS-srcinfo.nb_bits)>0) // left shift
      { for (i = 0; i < ns; i++)
          PasteBuf[i] = PasteBuf[i] << nshift;
      }
      // shift et swap du destinataire
      if ((nshift=NB_BITS-dstinfo.nb_bits)>0) // right shift
      { for (i = 0; i < ns; i++)
          PasteBuf[i] = PasteBuf[i] >> nshift;
      }
      if (dstinfo.swap!=SWAPINFILE)
	swap((char *) &PasteBuf[0] ,nb );

      if (erboxmmioWrite(15,proc,dstinfo.nom
		    ,nb=mmioWrite(dstinfo.hmmio
		          ,(HPSTR)&PasteBuf[0],nb))
                !=0) nbread=-1;
      else
      { nbread+=nb;
        bcount-=nb;
      }
    }
  }
  PasteBuf_.newsize(1);
  return nbread;
}


///////////////////////////////////////////////
// writeSoundFile est la classe  utiliser
//   pour crire un fichier son
//
// toutes les informations  crire dans un fichier
//   son sont fournies  un objet writeSoundFile,
//   et l'criture elle-mme est faite d'un seul coup
//   par la procdure flush()   
//
// la bonne manire de l'utiliser est:
//     crer un objet writeSoundFile
//     initialiser ses lments
//       d'entte: dstinfo dsignant le fichier destination
//                 avec les lments suivants
//                     .nom   (toujours)
//                     .frequence .nb_bits .swap
//                     (facultatifs dans certains
//                      cas - voir plus loin)
//       de fichier source
//               par srcfile()
//       de donnes mmoire
//               par addmemdata() ou addfiledata()
//       de donnes tiquetage
//               par des addetiq()
//       de donnes annotations
//               par des addnote() 
//     effectuer un et un seul flush() 
//     dtruire l'objet writeSoundFile
//
// remarques:
// - si le format du fichier dst n'est pas .WAV
//       alors l'en-tte est prise dans le fichier
//       src, mme si le fichier src n'est pas de
//       mme format (dans ce dernier cas, l'en-tte
//       a la bonne longueur, mais un contenu
//       folklorique).
// - la dclaration du fichier src par srcfile() peut
//   tre omise si on n'utilise pas de fichier src,
//   c'est--dire:
//          fichier dst de format .WAV
//       et 1re partie vide (length1==0)
//       et 3me partie vide (length3==0)
//

/////////////////////////////////////////
// classe writeSoundFile
//
writeSoundFile::writeSoundFile()
{ dstinfo=NULL;
  insertinfo=NULL;
  length1=0;
  length2=0;
  length3=0;
  //
  petiqsbuf= (char *) NULL;
  netiqsbuf=0;
  ietiqsbuf=0;
  //
  pnotesbuf= (char *) NULL;
  nnotesbuf=0;
  inotesbuf=0;
}
writeSoundFile::~writeSoundFile()
{
  if (petiqsbuf) {
  	 delete []petiqsbuf;
    petiqsbuf = (char *) NULL;
  }
  if (pnotesbuf) {
  	 delete []pnotesbuf;
    pnotesbuf = (char *) NULL;
  }
}
void writeSoundFile::srcfile(infoFile *srcinfo
		      , long offset1, long length1
		      , long offset3, long length3)
{
  this->srcinfo=srcinfo;
  this->offset1=offset1;
  this->length1=length1;
  this->offset3=offset3;
  this->length3=length3;
}
void writeSoundFile::addfiledata(infoFile *insertinfo
		      , long offset, long length)
{
  this->insertinfo = insertinfo;
  offset2=offset;
  length2=length;
}

void writeSoundFile::addmemdata(SAMPLEbuf& buf
		      , long offset, long length)
{ insertinfo=NULL;
  buf2=&buf;
  offset2=offset;
  length2=length;
}
void writeSoundFile::addetiq(long echt, char *text)
{ int ns;

  // calculer le nombre d'octets supplmentaires
  ns=4+strlen(text)+1;
  // et rallonger le buffer si ncessaire
  if ( (ietiqsbuf+ns+2)>=netiqsbuf )
  {
    char *x;
    int i;
    x = new char[netiqsbuf+100];
    for (i = 0 ; i < netiqsbuf; i++)  x[i] = petiqsbuf[i];
    delete []petiqsbuf;
    netiqsbuf+=100;
    petiqsbuf = x;
  }
  //
  *((long *)(petiqsbuf+ietiqsbuf))=echt; // placer la limite
  lstrcpy((char *)(petiqsbuf+ietiqsbuf+4),text); // placer le texte
  //
  previetiqsbuf=ietiqsbuf;
  ietiqsbuf+=ns;
}
void writeSoundFile::addnote(long echt, char *text)
{ int ns;

  // calculer le nombre d'octets supplmentaires
  ns=4+strlen(text)+1;
  // et rallonger le buffer si ncessaire
  if ( (inotesbuf+ns+2)>=nnotesbuf )
  {
    char *x;
    int i;
    x = new char[nnotesbuf+100];
    for (i = 0 ; i < nnotesbuf; i++)  x[i] = pnotesbuf[i];
    delete []pnotesbuf;
    nnotesbuf+=100;
    pnotesbuf = x;
  }
  //
  *((long *)(pnotesbuf+inotesbuf))=echt; // placer la limite
  lstrcpy((char *)(pnotesbuf+inotesbuf+4),text); // placer le texte
  //
  previnotesbuf=inotesbuf;
  inotesbuf+=ns;
}
int writeSoundFile::flushWAV()
// en retour 0 si ok
{ int ret=0;
  long lriff,ldata; // longueur  mettre
               // dans les chunks RIFF et data
  typedef struct header
  { char c1[4];
    long  l1;
    char c2[8];
    long l2;
    PCMWAVEFORMAT w; // waveFormat;
    char c3[4];
    long ldata;
  } HEADER;
  HEADER header
  // ce header va du 1er octet du fichier inclus
  //     jusqu'au 1er octet des datas exclus
    = { {'R','I','F','F'},
        0,
        {'W','A','V','E','f','m','t',' '},
        0x10,
	{ {WAVE_FORMAT_PCM,1,11025,2*11025,2} ,16},
	// les initialisations ci-dessus sont donnes
	//   pour l'exemple. Les 11025 & co sont
	//   rinitialiss en ralit plus loin.
        {'d','a','t','a'},
        0
      };
  char proc[]="writeSoundFile::flushWAV()";
  STRING dstname; // idem dstinfo->nom

  if (debug==2) Wmsg->wr(proc);

  // calcul des longueurs exactes
  //   des tiquettes et notes
  // et contrle (en fait, suppression du dernier
  // texte qui doit tre de longueur nulle)
  if (ietiqsbuf!=0)
  { if (ietiqsbuf-previetiqsbuf!=5)
      erboxmmio(41,proc,dstinfo->nom,ml(1201,"extra end label"));
    ietiqsbuf=previetiqsbuf+4;
  }
  if (inotesbuf!=0)
  { if (inotesbuf-previnotesbuf!=5)
      erboxmmio(42,proc,dstinfo->nom,ml(1202,"extra end note"));
    inotesbuf=previnotesbuf+4;
  }

  // calcul des longueurs des chunks
  ldata=length1+length2+length3; //0;
  lriff=     sizeof(HEADER) -8
	    +ldata    +(ldata    &1);
  if (ietiqsbuf!=0)
    lriff+= 8 +ietiqsbuf+(ietiqsbuf&1);
  if (inotesbuf!=0)
    lriff+= 8 +inotesbuf+(inotesbuf&1);

  header.ldata=ldata;
  header.l1=lriff;
  header.w.wf.nSamplesPerSec=dstinfo->frequence;
  header.w.wf.nAvgBytesPerSec=2*dstinfo->frequence;
  header.w.wf.nBlockAlign=2;
  header.w.wBitsPerSample=16;

  strcpy(dstname,dstinfo->nom);
  dstinfo->nb_bits=NB_BITS;
  dstinfo->swap=SWAPINFILE;
  if (erboxmmioOpen(1,proc,dstname
	,dstinfo->hmmio=mmioOpen(dstinfo->nom,NULL
		   ,MMIO_WRITE|MMIO_CREATE))
	!=0) return 1;

  if (ret==0)
  { if (erboxmmioWrite(2,proc,dstname
        ,mmioWrite(dstinfo->hmmio,(HPSTR)&header,sizeof(HEADER)))
	  !=0) ret=1;
  }

  // datas
  if (ret==0)
  { int indopen; 

    // ouverture fichier src
    indopen=0;
    if ((ret==0)&&((length1!=0)||(length3!=0)))
    {
      //MMIOINFO x;
      indopen=1;
      //memset((void *)(&x), 0, sizeof(MMIOINFO));
      srcinfo->hmmio=mmioOpen(srcinfo->nom,NULL,MMIO_READ);
      if (erboxmmioOpen(12,proc,srcinfo->nom,srcinfo->hmmio)
	    !=0) ret=1;
    }

    // 1re partie
    if ((ret==0)&&(length1!=0))
      copyHFileToHFile(*srcinfo,offset1 
	              ,*dstinfo,length1);
    // 2me partie
    if ((ret==0)&&(length2!=0)) {
     	if (insertinfo==NULL) { // memory
			copyBufToHFile(*buf2,offset2/sizeof(SAMPLE)
	              			,*dstinfo,length2);
      }
      else { // file
			if (erboxmmioOpen(111,proc,insertinfo->nom
				,insertinfo->hmmio=mmioOpen(insertinfo->nom
		     	,NULL,MMIO_READ))!=0) ret=1;
			else {
         	copyHFileToHFile(*insertinfo,offset2
			  							,*dstinfo,length2);
	  			if (erboxmmioClose(112,proc,insertinfo->nom
		    			,mmioClose(insertinfo->hmmio,0))!=0) ret=1;
			}
      }
    }

    // 3me partie
    if ((ret==0)&&(length3!=0))
      copyHFileToHFile(*srcinfo,offset3
	              ,*dstinfo,length3);

    // fermeture fichier src
    if ((ret==0)&&(indopen==1))
      erboxmmioClose(16,proc,srcinfo->nom
	      ,mmioClose(srcinfo->hmmio,0));

    if ((ldata&1)!=0) erboxmmioWrite(17,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,"",1));
  }

  // etiquetage phontique
  if ((ret==0)&&(ietiqsbuf!=0))
  { erboxmmioWrite(21,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,"phon",4));
    erboxmmioWrite(22,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,(HPSTR)&ietiqsbuf,4));
    erboxmmioWrite(23,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,(HPSTR)petiqsbuf
	                       ,ietiqsbuf));
    if ((ietiqsbuf&1)!=0)
      erboxmmioWrite(24,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,"",1));
  }

  // annotations
  if ((ret==0)&&(inotesbuf!=0))
  { erboxmmioWrite(31,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,"word",4));
    erboxmmioWrite(32,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,(HPSTR)&inotesbuf,4));
    erboxmmioWrite(33,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,(HPSTR)pnotesbuf
	                       ,inotesbuf));
    if ((inotesbuf&1)!=0)
      erboxmmioWrite(34,proc,dstname
	       ,mmioWrite(dstinfo->hmmio,"",1));
  }

  if (erboxmmioClose(41,proc,dstname
		    ,mmioClose(dstinfo->hmmio,0))
	!=0) ret=1;

  return ret;
}
int writeSoundFile::flushNOTWAV()
// en retour 0 si ok
{ int ret=0;
  STRING dstname;
  char proc[]="writeSoundFile.flushNOTWAV()";

  if (debug==2) Wmsg->wr(proc);

  strcpy(dstname,dstinfo->nom);
  if (erboxmmioOpen(1,proc,dstname
	,dstinfo->hmmio=mmioOpen(dstinfo->nom,NULL
		   ,MMIO_WRITE|MMIO_CREATE))
	!=0) return 1;

  // datas
  if (ret==0)
  { int indopen;

    // ouverture fichier src
    indopen=0;
    if (ret==0)
    { indopen=1;
      if (erboxmmioOpen(12,proc,srcinfo->nom 
		,srcinfo->hmmio=mmioOpen(srcinfo->nom 
		     ,NULL,MMIO_READ))
	    !=0) ret=1;
    }

    // entete
    if (ret==0)
      copyHFileToHFile(*srcinfo,0,*dstinfo,dstinfo->depl);

    // 1re partie
    if ((ret==0)&&(length1!=0))
      copyHFileToHFile(*srcinfo,offset1,*dstinfo,length1);

    // 2me partie
    if ((ret==0)&&(length2!=0))
    { if (insertinfo==NULL)
      { // memory
	copyBufToHFile(*buf2,offset2/sizeof(SAMPLE)
	              ,*dstinfo,length2);
      }
      else
      { // file
	if (erboxmmioOpen(111,proc,insertinfo->nom
		,insertinfo->hmmio=mmioOpen(insertinfo->nom
		     ,NULL,MMIO_READ))
	    !=0) ret=1;
	else
	{ copyHFileToHFile(*insertinfo,offset2
			  ,*dstinfo,length2);
	  if (erboxmmioClose(112,proc,insertinfo->nom
		    ,mmioClose(insertinfo->hmmio,0))
              !=0) ret=1;
	}
      }
    }

    // 3me partie
    if ((ret==0)&&(length3!=0))
      copyHFileToHFile(*srcinfo,offset3 
	              ,*dstinfo,length3);

    // fermeture fichier src
    if ((ret==0)&&(indopen==1))
      erboxmmioClose(16,proc,srcinfo->nom 
	      ,mmioClose(srcinfo->hmmio,0));
  }

  if (erboxmmioClose(41,proc,dstname
		    ,mmioClose(dstinfo->hmmio,0))
	!=0) ret=1;

  return ret;
}
int writeSoundFile::flush()
// en retour 0 si ok
{ if (debug==2)
  { STRING ch;
    Wmsg->wr("writeSoundFile::flush()");
    sprintf(ch,"... %s",dstinfo->nom);
    Wmsg->wr(ch);
  }
  if (isWAV(dstinfo->nom)) return flushWAV();  
  return flushNOTWAV();
}


/*
 * Copier une rgion dans le cut file
 */
void copieDansCutFile(struct editeur *infoEdit)
{ long nbSample; 
  int i;

  writeSoundFile wS;
  notetiqs notetiqscut;

    nbSample = infoEdit->fin -  infoEdit->debut + 1;

    wS.dstinfo=&infoEdit->cut;
    //
    wS.addmemdata(space,_i(infoEdit->debut)*sizeof(SAMPLE)
          ,nbSample*sizeof(SAMPLE) );
    etiqs.edit_copy(infoEdit->debut,infoEdit->fin
              ,notetiqscut);
    for (i=0;i<notetiqscut.nb;i++)
      wS.addetiq(notetiqscut.e[i]->limite
                ,notetiqscut.e[i]->Text());
    notetiqscut.clear();
    notes.edit_copy(infoEdit->debut,infoEdit->fin
              ,notetiqscut);
    for (i=0;i<notetiqscut.nb;i++)
      wS.addnote(notetiqscut.e[i]->limite
                ,notetiqscut.e[i]->Text());
    //
    wS.flush();
    //

  // est-ce bien utile ? normalement non
  infoEdit->cut.depl = 0;
  infoEdit->cut.longueur=nbSample*sizeof(SAMPLE);

}

/* repercuteModif()
 * Rpercution des modifications du buffer dans
 *   le fichier de parole (en fait sauvegarde).
 */
long repercute_tf()
// renvoie le numro du dernier chantillon  sauver
//   c'est--dire le futur en_cours.tf-1
{ return ct.f_e + ct.tf - infoEdit.derEchtFic -1;
}
void repercuteModif(
	struct editeur *infoEdit)
// sauve dans le fichier infoEdit->save
//   les informations courantes constitues par
//   le fichier courant et le tampon courant,
// les contrles de droit d'accs en criture
//   sur infoEdit->save sont dj faits
// les tiquetages ne sont pas sauvs, et doivent donc
//   l'tre par l'appelant.
// en fin, infoEdit->save devient le nouveau fichier courant
// si infoEdit->save est identique au fichier courant, alors
//   le fichier courant initial est renomm en *.bak
//
// en retour, TRUE ok
{
  int i;
  long il;
  writeSoundFile wS;
  infoFile srcinfo; // fichier source

  srcinfo=en_cours;
  if (strcmp(infoEdit->save.nom,en_cours.nom)==0) {
   { char dr[_MAX_DRIVE],dir[_MAX_DIR],name[_MAX_FNAME],ext[_MAX_EXT];
      _splitpath(en_cours.nom,dr,dir,name,ext);
      strcpy(ext,".bak");
      _makepath(srcinfo.nom,dr,dir,name,ext);
    }
    remove(srcinfo.nom);
    rename(en_cours.nom,srcinfo.nom);
  }
  else strcpy(srcinfo.nom,en_cours.nom);

  wS.dstinfo = &infoEdit->save;
  il = (infoEdit->derEchtFic + 1) * sizeof(SAMPLE);
  wS.srcfile(&srcinfo, en_cours.depl, ct.d_e * sizeof(SAMPLE), en_cours.depl + il, en_cours.longueur - il );
  wS.addmemdata(space, _i(ct.d_e) * sizeof(SAMPLE), (ct.f_e - ct.d_e + 1) * sizeof(SAMPLE));
  //
  for (i=0; i < etiqs.nb && etiqs.e[i]->limite < ct.tf;i++)
    wS.addetiq(etiqs.e[i]->limite,etiqs.e[i]->Text());
  //
  for (i=0; i < notes.nb && notes.e[i]->limite < ct.tf;i++)
    wS.addnote(notes.e[i]->limite,notes.e[i]->Text());
  //

  wS.flush();
  infoEdit->save.mode = en_cours.mode;
  en_cours = infoEdit->save;
  init_infoFile_read(&en_cours);
  ct.tf = en_cours.longueur/sizeof(SAMPLE);

  infoEdit->derEchtFic = ct.f_e;
  infoEdit->yaDesModifsSignal = FAUX;
  infoEdit->yaDesModifsEtiquetages = FAUX;
}

//
void Set_yaDesModifsSignal()
{ infoEdit.yaDesModifsSignal=TRUE;
  pm.clear();
  piqueforClear();
  trajOLAClear();
}
void Set_yaDesModifsEtiquetages()
{ infoEdit.yaDesModifsEtiquetages=TRUE;
}

void initEditSignalSeul()
{
  infoEdit.derEchtFic = ct.f_e;
  infoEdit.yaDesModifsSignal = FAUX;
  infoEdit.yaDesModifsEtiquetages = FAUX;
  infoEdit.derOp = PASDOP;
  ct.debRegion = 0;
  ct.finRegion = 0;
  ct.posBarre = -1;
}

void initEditSignal()
{
  initEditSignalSeul();
  pm.clear();
  piqueforClear();
  trajOLAClear();
}


////////////////////////////////////////////
// insertFileIntoCurrentFile()
//
//   insertion d'un morceau de fichier dans le
//     fichier courant de parole
//
// le fichier  insrer est donn par infoEdit->insert
//                                 et infoEdit->debut
// le fichier courant est donn par en_cours
//
void insertFileIntoCurrentFile(struct editeur *infoEdit)
// insre un morceau de fichier parole
//     dans le fichier courant de parole
// les annotations et tiquetage sont
//   dj  jour. Il reste juste  les mettre
//   aussi dans le fichier insert.
{ int i;
  long il;
  writeSoundFile wS;
  infoFile srcinfo;

    wS.dstinfo=&en_cours;
    srcinfo=en_cours;
    // renommer le fichier courant en *.bak
    { char dr[_MAX_DRIVE],dir[_MAX_DIR],name[_MAX_FNAME],ext[_MAX_EXT];
      _splitpath(en_cours.nom,dr,dir,name,ext);
      strcpy(ext,".bak");
      _makepath(srcinfo.nom,dr,dir,name,ext);
    }
    remove(srcinfo.nom);
    rename(en_cours.nom,srcinfo.nom);
    //
    il= infoEdit->debut * sizeof(SAMPLE);
    wS.srcfile(&srcinfo
	       ,en_cours.depl
	       ,il 
	       ,en_cours.depl+il
	       ,en_cours.longueur-il );
    wS.addfiledata(&infoEdit->insert
		  ,infoEdit->insert.depl
		  ,infoEdit->insert.longueur);
    //
    { // etiquetages phontique et orthographique
      for (i=0;i<etiqs.nb;i++)
        wS.addetiq(etiqs.e[i]->limite
                ,etiqs.e[i]->Text());
      // 
      for (i=0;i<notes.nb;i++)
        wS.addnote(notes.e[i]->limite
		,notes.e[i]->Text());
    }
    //
    if (wS.flush()!=0)
    { STRING ch;
      sprintf(ch,"unexpected error in inserting\n\
file %s\nin file %s",infoEdit->insert.nom
		    ,en_cours.nom);
      affiche_bug(ch); 
    }
  else 
    // mettre  jour en_cours
    init_infoFile_read(&en_cours);
}

////////////////////////////////////////////
// substituteFileIntoCurrentFile()
//
//   substitution d'un morceau de fichier dans le
//     fichier courant de parole
//
// le fichier  substituer est donn par infoEdit->substitute
//                                 et infoEdit->debut
// le fichier courant est donn par en_cours
//
void substituteFileIntoCurrentFile(struct editeur *infoEdit)
// insre un morceau de fichier parole
//     dans le fichier courant de parole
// les annotations et tiquetage sont
//   dj  jour. Il reste juste  les mettre
//   aussi dans le fichier insert.
{ int i;
  long il;
  writeSoundFile wS;
  infoFile srcinfo;

    wS.dstinfo=&en_cours;
    srcinfo=en_cours;
    // renommer le fichier courant en *.bak
    { char dr[_MAX_DRIVE],dir[_MAX_DIR],name[_MAX_FNAME],ext[_MAX_EXT];
      _splitpath(en_cours.nom,dr,dir,name,ext);
      strcpy(ext,".bak");
      _makepath(srcinfo.nom,dr,dir,name,ext);
    }
    remove(srcinfo.nom);
    rename(en_cours.nom,srcinfo.nom);
    //
    il= infoEdit->debut * sizeof(SAMPLE);
    wS.srcfile(&srcinfo,
	            en_cours.depl,
	       	   il,
	       		en_cours.depl + il + infoEdit->substitute.longueur,
	            max(0, en_cours.longueur - il - infoEdit->substitute.longueur));
    wS.addfiledata(&infoEdit->substitute,
		  				 infoEdit->substitute.depl,
		  				 infoEdit->substitute.longueur);
    //  Phonetic and orthographic annotations are not considered !
    if (wS.flush()!=0) {
      STRING ch;
      sprintf(ch,"unexpected error in inserting\n\
file %s\nin file %s",infoEdit->substitute.nom, en_cours.nom);
      affiche_bug(ch); 
    }
    else // update en_cours
    init_infoFile_read(&en_cours);
}

/*
 * Sauve une region de la mmoire vers un fichier
 */
void sauveRegion(struct editeur *infoEdit)
// sauve la region (infoEdit->debut  infoEdit->fin inclus)
// dans le fichier infoEdit->save.nom
// en utilisant infoEdit->save.nb_bits et infoEdit->save.swap
// et met  jour infoEdit->save.longueur
{ int i;
  long il;
  writeSoundFile wS;
  notetiqs nesave;

  wS.dstinfo=&infoEdit->save;
  //
  il= (infoEdit->fin-infoEdit->debut+1)*sizeof(SAMPLE);
  // utile uniquement dans le cas fichier non .WAV
  wS.srcfile(&en_cours
	       ,en_cours.depl
	       ,0,0,0 );
  //
  wS.addmemdata(space,_i(infoEdit->debut)*sizeof(SAMPLE)
          ,il); //(ct.f_e - ct.d_e + 1)*sizeof(SAMPLE) );
  //
  etiqs.edit_copy(infoEdit->debut,infoEdit->fin,nesave);
  if (isWAV(infoEdit->save.nom))
    for (i=0;i<nesave.nb;i++)
      wS.addetiq(nesave.e[i]->limite,nesave.e[i]->Text());
  else
  { if (infoEdit->save.fmt_phon!=0)
    { if ((isEtiqModifiable
	   =droit_ecrire_message(infoEdit->save.nom_man)
          )==TRUE)
        sauvman(&nesave, infoEdit->save);
    }
  }
  //
  nesave.clear();
  notes.edit_copy(infoEdit->debut,infoEdit->fin,nesave);
  if (isWAV(infoEdit->save.nom))
    for (i=0;i<nesave.nb;i++)
      wS.addnote(nesave.e[i]->limite,nesave.e[i]->Text());
  else
  { if (infoEdit->save.fmt_word!=0)
    { if ((isNoteModifiable
	   =droit_ecrire_message(infoEdit->save.nomAnnot)
	  )==TRUE)
	sauveAnnotations(&nesave, infoEdit->save);
    }
  }

  //
  wS.flush();

  if (!isWAV(infoEdit->save.nom))
  { if (  (infoEdit->save.fmt_phon!=0)
        &&(isExt(infoEdit->save.nom_man)) )
      sauvman(&etiqs, infoEdit->save);
    if (  (infoEdit->save.fmt_word!=0) 
	&&(isExt(infoEdit->save.nomAnnot)) )
      sauveAnnotations(&notes, infoEdit->save);
  }
}


//
// Lit un morceau du fichier de parole
//
// retourne le nombre d'chantillons lus
long lireSignal(
	infoFile& srcinfo, // fichier source 
	long srcoffset,	 // offset exprim en # de caractres
	                 //    partir de dbut du fichier 
	SAMPLEbuf& dest, // destination
	long dest_debut, // dbut dans le buffer dest
	long nbEcht)	 // nombre d'chantillons  lire
// en retour, le nombre de SAMPLEs lus
//
// normalement les SAMPLEs demands doivent tre
// disponibles dans le fichier,
// car l'appelant en a dtermin auparavant le nombre.
// Si ce n'est pas le cas, le nombre de SAMPLEs lus
// donn en retour est infrieur au nombre de SAMPLEs
// demands, (aucune erreur n'est signale).
//
// NB:
// - dans srcinfo sont utiliss
//       nom, nb_bits, swap
{ int ret=0;
  long nbEchtLu;
  long reelOffset;
  char proc[]="lireSignal()";

  // contrle des bornes de la lecture
  //     par rapport au buffer destinataire
  if (   (dest_debut < 0)
      || (nbEcht < 0)
      || (dest_debut + nbEcht > dest.size())) {
    affiche_bug("lireSignal() : try to read data out of the buffer");
    return 0;
  }

  if (erboxmmioOpen(1,proc,srcinfo.nom
    	,srcinfo.hmmio=mmioOpen(srcinfo.nom,NULL,MMIO_READ))
	!=0) return 0;

    reelOffset=mmioSeek(srcinfo.hmmio,srcoffset,SEEK_SET);
    if (reelOffset != srcoffset) {
      affiche_bug("lireSignal(): offset out of file");
      nbEchtLu=0;
      ret=1;
    }
    if (ret==0) {
      long nb_lus;
      nb_lus=copyHFileToBuf(srcinfo,srcoffset
                         ,dest,dest_debut
			 ,nbEcht*sizeof(SAMPLE));
      if (nb_lus==-1) nbEchtLu=0;
      else nbEchtLu=nb_lus/sizeof(SAMPLE);
    }

  erboxmmioClose(4,proc,srcinfo.nom
		     ,mmioClose(srcinfo.hmmio,0));

  return(nbEchtLu);
}

void initedit()
{ strcpy(infoEdit.insert.nom,"");
  infoEdit.insert.depl=0;
  infoEdit.insert.nb_bits=0;
  infoEdit.insert.swap=0;
  //
  infoEdit.cut.depl=0;
  infoEdit.cut.nb_bits=NB_BITS;
  infoEdit.cut.swap=SWAPINFILE;
  //
  strcpy(infoEdit.save.nom,"");
  infoEdit.save.depl=0;
  infoEdit.save.nb_bits=0;
  infoEdit.save.swap=0;
  //
  infoEdit.derOp=PASDOP;
  infoEdit.debut=-1;
  infoEdit.fin=-1;
  infoEdit.derEchtFic=-1;
  infoEdit.yaDesModifsSignal=FALSE;
  infoEdit.yaDesModifsEtiquetages=FALSE;
}
