/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

// fichier fftproto.h
//

#ifndef _FFTPROTO_H
#define _FFTPROTO_H

#include <stdio.h>
#include <malloc.h>
#include <math.h>

/*=================================================================*
 * Prototype de cacul de FFT complexe				   *
 * IL FAUT definir les constantes :				   *
 * N : l'ordre de la FFT					   *
 * LOG2N : le log a base 2 de N					   *
 * ENTETE le nom de la fonction engendree ainsi que ses parametres *
 * ENTREE la maniere de traduire les donnees en entree pour la FFT *
 *	  Il FAUT remplir les tableaux x et y a partir de 1 avec   *
 *	  exactement N valeurs					   *
 * TYPE   le type des calculs internes a la FFT (double par        *
 *	  exemple)						   *
 * SORTIE la maniere de retourner les resultats (voir au-dessous)  *
 *        la variable j1 est disponible pour stocker tabjn_[]      *
 *        ainsi que t1 et t2 (float)				   *
 * FFTINIT le nom de la fonction d'initialisation des tables 	   *
 *	   auxiliaires. Pour engendrer une FFT directe il suffit   *
 *	   d'appeler cette fonction avec 1, pour une FFT directe   *
 *	   il faut l'appeler avec -1. Les resultas ne sont alors   *
 *	   pas divises par N.					   *
 * LIBERE  le nom de la fonction de supression de l'espace des     *
 *	   tables annexes				           *
 * MESFFT  chaine de caracteres donnant le nom de la fft           *
 * MESINIT chaine de caracteres donnant le nom de l'initialisation *
 * MESFREE chaine de caracteres du nom de la fct de destruction    *
 * exemple d'appel :						   *
 *   initfft(1);						   *
 *   fft(data);							   *
 * NOTE : Ce prototype est destine a construire une machine FFT    *
 *	  d'un ordre donne et qui doit etre utilisee un grand      *
 *	  nombre de fois ; par consequent il y a autant de         *
 *	  repliques du code qu'il y a de types de FFT              *
 *=================================================================*/

// N doit tre au plus gal  1024, pour la taille
// des tableaux. Pour des valeurs suprieures,
// il faut revrifier


#define PI    3.14159265358979323846
#define PIX2  6.28318530717958647692

TabAuxFFT AUX = {(int *) NULL, (int *) NULL, (int *) NULL,
						(TYPE *) NULL, (TYPE *) NULL, false};
static void FFTINIT(int isign);

/* entete de la fonction calculant la fft
 * par exemple :
 * #define ENTREE fftExemple(data) float data[];
 */
ENTETE
{
  register i,j1,j2;
  TYPE x[N + 1];
  TYPE y[N + 1];
  TYPE t1,t2;

  if (!AUX.init) { AUX.init = true; FFTINIT(1); }

  /* entree des donnees et transfert dans les tableaux x et y
   * en commencant a l'indice 1.
   * Par exemple
   * #define ENTREE   for (i = 1; i  <= N; i++) {       \
   * 		          x[i] = data[2*i-1]; 	        \
   * 		          y[i] = data[2*i];  	        \
   *	    	      }
   */

  ENTREE

  for (i = 1; i <= (N / 2) * LOG2N; i++) {
    j1 = AUX.tabj1_[i];
    j2 = AUX.tabj2_[i];
    t1 = x[j1] - x[j2];
    t2 = y[j1] - y[j2];
    x[j1] = x[j1] + x[j2];
    y[j1] = y[j1] + y[j2];
    x[j2] = AUX.tabcos_[i] * t1 + AUX.tabsin_[i] * t2;
    y[j2] = AUX.tabcos_[i] * t2 - AUX.tabsin_[i] * t1;
  }

  /* desembrouillage et tranfert de sortie.
   * Le tableau de desembrouillage est tabjn_
   * en commencant a l'indice 1.
   *  Par exemple :
   *  #define SORTIE     for (i = 1; i <= N; i++) { \
   * 		           j1 = tabjn_[i];	    \
   *                       data[i + i - 1] = x[j1]; \
   *	                   data[i + i] = y[j1];     \
   *	                 }
   */

  SORTIE
}

/*
 *
 */

static void FFTINIT(int isign)

{
	int p2[LOG2N + 1];
   	int i,lo,k;
   	int lmx;
   	int lix;
   	double scl;
   	int lm;
   	double arg;
   	double c,s;
   	int li;

	if (AUX.tabj1_ != NULL || AUX.tabj2_ != NULL || AUX.tabcos_ != NULL || AUX.tabsin_ != NULL
	    || AUX.tabjn_ != NULL) {
	   fprintf(stderr, "%s %s", MESINIT, ": tables auxiliaires deja allouees\n");
	   return;
	}
	AUX.tabj1_ = (int *) malloc((unsigned int) (((N / 2) * LOG2N + 1) * sizeof(int)));
	AUX.tabj2_ = (int *) malloc((unsigned int) (((N / 2) * LOG2N + 1) * sizeof(int)));
	AUX.tabcos_ = (TYPE *) malloc((unsigned int) (((N / 2) * LOG2N + 1) * sizeof(TYPE)));
	AUX.tabsin_ = (TYPE *) malloc((unsigned int) (((N / 2) * LOG2N + 1) * sizeof(TYPE)));
	AUX.tabjn_ = (int *) malloc((unsigned int) ((N + 1) * sizeof(int)));

	initMask(LOG2N);
	
	for (i = 1, p2[0] = 1; i<= LOG2N; i++) p2[i] = p2[i-1] << 1;

	for (lo = 1, k = 1; lo <= LOG2N; lo++) {
       	    lmx = p2[LOG2N - lo];
            lix = 2 * lmx;
            scl = (double) PIX2 / lix;
            for (lm = 1; lm <= lmx; lm++) {
                arg = (double) scl * (lm - 1) * isign;
                c = (TYPE) cos(arg);
                s = (TYPE) sin(arg);
                li = lix;
                while (li <= N) {
                      AUX.tabj1_[k] = li - lix + lm;
                      AUX.tabj2_[k] = AUX.tabj1_[k] + lmx;
                      AUX.tabcos_[k] = c;
                      AUX.tabsin_[k] = s;
                      k++;
                      li += lix;
		}
	    }
	}

	for (i = 1; i <= N ; i++) AUX.tabjn_[i] = bitReverseFft(i, LOG2N);
}

/*
 * Cette fonction libere l'expace auxiliaire necessaire au calcul de la
 * fft.
 */

LIBERE
{
	if (AUX.tabj1_ == NULL || AUX.tabj2_ == NULL || AUX.tabcos_ == NULL || AUX.tabsin_ == NULL
	    || AUX.tabjn_ == NULL) {
	   fprintf(stderr, "%s %s", MESFREE, ": pas de tables auxiliaires\n");
	   return;
	}
	free((char *) AUX.tabj1_);
	AUX.tabj1_ = NULL;
	free((char *) AUX.tabj2_);
	AUX.tabj2_ = NULL;
	free((char *) AUX.tabcos_);
	AUX.tabcos_ = NULL;
	free((char *) AUX.tabsin_);
	AUX.tabsin_ = NULL;
	free((char *) AUX.tabjn_);
	AUX.tabjn_ = NULL;
}

#undef ENTETE
#undef ENTREE
#undef SORTIE
#undef TYPE
#undef N
#undef LOG2N
#undef AUX
#undef LIBERE
#undef MESFFT
#undef MESINIT
#undef MESFREE
#undef FFTINIT

#endif






