/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

/********************************************************************/
/*     Programme ecrit a l'aide du document suivant :               */
/*        Techniques cepstrales ameliorees pour l'extraction        */
/*          d'enveloppe spectrale et la detection du pitch          */
/*      Par Pierre HALLE                                            */
/*      Pour :                                                      */
/*             Actes du seminaire "Traitement du signal de Parole"  */
/*                      ENST, Paris, 15-16 decembre 1983            */
/********************************************************************/


#include <math.h>
#include <stdio.h>
#include "common.h"
#include "fftlpc.h"

#define PI    		 3.14159265358979323846
#define PIX2       6.2831853071795864769
#define N          4096
#define NBITERMAX  1

#define g(x) (((x)>0.)?(x):0.) 

static double piX2_N = (double) (PIX2 / N);
static void initfiltre(int n1, int n2);
static bool premFois = true;
static int hammingWindowSize;
static FLOATbuf hamming(N,"hamming");   /* fenetre de Hamming          */
static FLOATbuf liftre(N, "liftre");
static FLOATbuf S(N,"S"),         /* Sk = ln (|Xk|^2)            */
        c(2*N,"c"),		/* c = IDFT( S )                  */
        C(N,"C"),		/* Cm = hm * cm                   */
        S_(2*N,"S_"),        /* S_ = DFT( C )                  */
        Vi(N,"Vi"),		/* lissage cepstral               */
        Viplus1(N,"Viplus1"),	/* iteration suivante             */
        Ei(N,"Ei"),		/* depassement                    */
        Eiplus1(N,"Eiplus1"),	/* depassement suivant            */
        E_i(N, "E_i"),		/* lissage du depassement         */
        e_i(2*N, "e_i"),       /* cepstre de E_i                 */
        E_iplus1(N, "E_iplus1"),	/* lissage du depassement suivant */
        tabtemp(2*N, "tabtemp");

/*
 *  Fonction qui calcule le lissage cepstral,
 *  pour le calcul propose par Halle.
 */

static void liss_cepstral(
  FLOATbuf &h,      /* definit la fenetre cepstrale */
  FLOATbuf &Ei,     /* donnee pour l'operation      */
  FLOATbuf &E_i,    /* resultat de l'operation      */
  int ordre)
{
  int k;
  FLOATbuf e_i(2*N, "e_i"); /* cepstre de Ei */

  /* initialisation de e_i */
  for (k=0;  k < ordre; k++) {
    e_i[2*k] = Ei[k];
    e_i[2*k+1] = 0. ;
  }

  switch (ordre) {
  case 256: fftic256 (e_i);
    break;
  case 512: fftic512 (e_i);
    break;
  case 1024: fftic1024 (e_i);
    break;
  case 2048: fftic2048 (e_i);
    break;
  case 4096: fftic4096 (e_i);
    break;
  }
 
  for (k=0; k<ordre; k++) {
    e_i[2*k] = (e_i[2*k] * liftre[k] * ((k) ? 2. : 1.)); 
    e_i[2*k+1] = 0. ;
  }

  switch (ordre) {
  case 256: fftc256 (e_i);
    break;
  case 512: fftc512 (e_i);
    break;
  case 1024: fftc1024 (e_i);
    break;
  case 2048: fftc2048 (e_i);
    break;
  case 4096: fftc4096 (e_i);
    break;
  }

  for (k=0; k<ordre; k++) E_i[k] = e_i[2 * k];
}


/*
 *  A partir de S, on va calculer les differents 
 *  lissages cepstraux, puis les depassements de
 *  ceux-ci sur les precedents : 
 *  algorithme iteratif
 */

void halle(
  float res[],   /* spectre lisse cepstralement. */
  SAMPLE cnv[],  /* signal de parole  */
  int debut,       // first sample of the signal window
  int ordreFFT,    // FFT order
  int fenSignal,   // # of samples in the window
  int n1, 
  int n2,	/* points caracteristiques du liftrage                */
  float alpha,  /* le coefficient accelerateur de convergence         */
  int nb_iter,	/* nombre d'iterations pour les calculs de la courbe  */
  bool preac)	/* true si preaccentuation. */
{
  int k, m, nbIter = 0;
  float temp;

  if (premFois || hammingWindowSize != fenSignal) {
    hammingWindowSize = fenSignal;
    initham(&hamming, fenSignal);
    premFois = false;
  }

  initfiltre (n1, n2);

  if (preac) for (k=0, m = 0; k < fenSignal; k++, m=m+2) {
    tabtemp[m] = (cnv[debut+k+1] - cnv[debut+k]) * hamming[k];
    tabtemp[m+1] = 0.;
  }  
  else for (k=0, m=0; k < fenSignal; k++, m=m+2) {
    tabtemp[m] = cnv[debut+k] * hamming[k];
    tabtemp[m+1] = 0.;
  }
  for (k = fenSignal, m = fenSignal * 2; k < ordreFFT; k++, m+=2) {
    tabtemp[m] = 0.;
    tabtemp[m + 1] = 0.;
  }

  /* tabtemp = DFT( x )   */
  switch (ordreFFT) {
  case 256: fftc256 (tabtemp);
    break;
  case 512: fftc512 (tabtemp);
    break;
  case 1024: fftc1024 (tabtemp);
    break;
  case 2048: fftc2048 (tabtemp);
    break;
  case 4096: fftc4096 (tabtemp);
    break;
  }

  /* S[k] = 10 * log10 ( x[k] * x[k] )     */
  /* comme c'est une suite symetrique,
     on ne calcule que la premiere partie  */
  for (k=0; k <= ordreFFT / 2; k++ ) {
     temp = tabtemp[2*k] * tabtemp[2*k] + tabtemp[2*k+1] * tabtemp[2*k+1];
     temp = (temp < EPSILON) ? EPSILON : temp;

     S[k] = 10. * log10 ( temp );
  }

  /* On rend le spectre symetrique. */
  for (k = 1; k <= ordreFFT / 2; k++)
    S[k + ordreFFT / 2] = S[ordreFFT / 2 - k];

  /* c = (c0, ... , cN-1) = IDFT(S) */
  for (k = 0; k < ordreFFT; k++) {
    c[2*k] = S[k];
    c[2*k+1] = 0.;
  }

  switch (ordreFFT) {
  case 256: fftic256 (c);
    break;
  case 512: fftic512 (c);
    break;
  case 1024: fftic1024 (c);
    break;
  case 2048: fftic2048 (c);
    break;
  case 4096: fftic4096 (c);
    break;
  }

  /* liftrage puis S^ = (S0, ... , SN-1) = DFT(C) */
  for (k = 0; k < ordreFFT; k++) {
    S_[2*k] = c[2*k] * liftre[k] * ((k) ? 2. : 1.) ;
    S_[2*k+1] = 0. ;
  }

  switch (ordreFFT) {
  case 256: fftc256 (S_);
    break;
  case 512: fftc512 (S_);
    break;
  case 1024: fftc1024 (S_);
    break;
  case 2048: fftc2048 (S_);
    break;
  case 4096: fftc4096 (S_);
    break;
  }

  /* Calcul des premiers elements de l'iteration */
  for (k = 0; k < ordreFFT; k++) {
    Vi[k] = S_[2*k] ;
    Ei[k] = g ( S[k] - S_[2*k] );
  }

  /* initialisation de e_i */
  for (k = 0; k < ordreFFT; k++) {
    e_i[2*k] = Ei[k] ;
    e_i[2*k+1] = 0. ;
  }
  switch (ordreFFT) {
  case 256: fftic256 (e_i);
    break;
  case 512: fftic512 (e_i);
    break;
  case 1024: fftic1024 (e_i);
    break;
  case 2048: fftic2048 (e_i);
    break;
  case 4096: fftic4096 (e_i);
    break;
  }

  /* Calcul du lissage cepstral du depassement */
  for (k = 0; k < ordreFFT; k++) {
    e_i[2*k] = e_i[2*k] * liftre[k] * ((k) ? 2. : 1.);
    e_i[2*k + 1] = 0.;
  }			
  switch (ordreFFT) {
  case 256: fftc256 (e_i);
    break;
  case 512: fftc512 (e_i);
    break;
  case 1024: fftc1024 (e_i);
    break;
  case 2048: fftc2048 (e_i);
    break;
  case 4096: fftc4096 (e_i);
    break;
  }
  for (k=0; k<ordreFFT; k++) E_i[k] = e_i[2*k];

  while (nbIter < nb_iter) {
    for (k=0; k<ordreFFT; k++) {
      Viplus1[k] = Vi[k] + E_i[k];
      Eiplus1[k] = g( Ei[k] - (1.+alpha) * E_i[k] );
    }

    liss_cepstral(liftre, Eiplus1, E_iplus1, ordreFFT);

    for (k=0; k<ordreFFT; k++) {
      Vi[k] = Viplus1[k];
      Ei[k] = Eiplus1[k];
      E_i[k] = E_iplus1[k];
    }
    nbIter++;
  }

  for (k = 0; k < ordreFFT / 2; k++) 
       res[k] = Vi[k];
}


/*
 * Initialisation du filtre cepstral.
 */
static void initfiltre(
  int n1, 
  int n2) 
{
  int i;

  for (i = 0; i < n1; i++) liftre[i] = 1.;

  for (i = n1; i < n2; i++) 
    liftre[i] = (float) 0.5 * (1. + cos ((double)(PI * (i - n1 + 1) / (n2 - n1))));

  for (i = n2; i < N; i++) liftre[i] = 0.;
}


#undef N
#undef PIX2
#undef NBITERMAX

