/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

#include "fft_mayer.h"
#include <math.h>
#include <vector>
#include "phase_unwrapping.h"

#ifndef M_PI
#define M_PI _PI
#endif

#define TWO_PI (M_PI + M_PI)

#ifdef DEBUG_unwrapping
void bruteForceRealFT(
  double x[],
  double *real,
  double *imaginary,
  int n,
  float omega) // normalized frequency (between 0 and 1)
{
   *real = *imaginary = 0.;
	for (int i = 0; i < n; i++) {
     *real += x[i] * cos(TWO_PI * omega * i);
     *imaginary += x[i] * sin(-TWO_PI * omega * i);
   }
}

void bruteForceFT(
  double x[],
  double y[],
  double *real,
  double *imaginary,
  int n,
  float omega) // between 0 and 2 * Pi)
{
   *real = *imaginary = 0.;
	for (int i = 0; i < n; i++) {
     *real += x[i] * cos(omega * i) - y[i] * sin(-omega * i);
     *imaginary += x[i] * sin(-omega * i) + y[i] * cos(omega * i);
   }
}
#endif

void realGoertzel(
  double x[],
  double *real,
  double *imaginary,
  int n,
  float omega) // frequency (between 0 and 2 * Pi)
{
  double fr = omega,
         cosa = cos(fr),
         sina = -sin(fr),
         cosb = cos(fr * (n - 1)),
         sinb = -sin(fr * (n - 1)),
         coeff = 2. * cosa,
         s_prev = 0,
         s_prev2 = 0,
	 tmp;
  int i;
  for (i = 0; i < n; i++) {
    double s = x[i] + coeff * s_prev - s_prev2;
    s_prev2 = s_prev;
    s_prev = s;
  }
  tmp = s_prev - cosa * s_prev2;
  *real = tmp * cosb + sina * sinb * s_prev2;
  *imaginary = tmp * sinb - sina * cosb * s_prev2; // diffrence de signe avec fft_mayer (oppose de la partie imaginaire)
}



#define ARGDerTHRESHOLD2 0.05 // less than PI

// Inspired by the paper of Tribolet (IEEE TASSP - A new phase unwrapping algorithm)
// see also the master thesis of Zahi Nadim Karam Computation of the one-dimensionnal unwrapped phase (MIT 2006)


/*
Scilab program to construct an artificial test signal, and to evaluate its phase and phase derivative.
1024 real points (the imaginary part of the signal is zero).
// without Pi jump
spectre= (1 - 0.95 * exp(%i * %pi *(1/4 - 2 *(0:1:1023) /1024))) .* (1 - 0.95 * exp(%i * %pi *(-1/4 - 2 *(0:1:1023) / 1024))) .* (1 - 0.99 * exp(%i * %pi *(2/3 - 2 *(0:1:1023) /1024))) .* (1 - 0.99 * exp(%i * %pi *(-2/3 - 2 *(0:1:1023) /1024)));
// with a Pi jump
spectre= (1 - 0.95 * exp(%i * %pi *(1/4 - 2 *(0:1:1023) /1024))) .* (1 - 0.95 * exp(%i * %pi *(-1/4 - 2 *(0:1:1023) / 1024))) .* (1 - exp(%i * %pi *(2/3 - 2 *(0:1:1023) /1024))) .* (1 - exp(%i * %pi *(-2/3 - 2 *(0:1:1023) /1024)));

signal = ifft(spectre);
nsignal = signal .* (0:1:1023);
plot2d(atan(imag(fft(signal)), real(fft(signal))));
write('signalPhaseLineSautPi',real(signal));
X = fft(signal);
Xp = fft(nsignal);

plot2d((real(X) .* real(Xp) + imag(X) .* imag(Xp)) .*abs(X)^(-2));
signal=read('signalDeuxPolesSautPi',-1,1024);
*/



void RealPhaseUnwrapping::removeLinearPhase()
{
    	for (int i = 0; i <= n / 2; i++) phase[i] -= phase[n / 2];
}

void RealPhaseUnwrapping::rearrange()
{
    	int i, j, nUnwrapping;
      nUnwrapping = ((int) standardFreq.size());
      for(i = 0, j = 0; i < nUnwrapping; i++) {
        double freq = standardFreq.at(i);
      	if (fabs(freq - floor(freq / dOmega + 0.5) * dOmega) < 1e-5)
            phase[j++] =  phaseUnwrapped.at(i);
      }
}

RealPhaseUnwrapping::~RealPhaseUnwrapping() {
      delete [] real;
      delete [] imag;
      delete [] realCopy;
      delete [] imagCopy;
      delete [] realnxn;
      delete [] imagnxn;
      delete [] realnxnCopy;
      delete [] imagnxnCopy;
      delete [] phaseDer;
      delete [] phase;
};

double *RealPhaseUnwrapping::get(double result[])
    {
      for(int i = 0; i <= n / 2; i++) result[i] =  phase[i];
      return(result);
    }

RealPhaseUnwrapping::RealPhaseUnwrapping(double *xr, int nbPoints, int fftOrder) {          // double *xi as the second argument if the signal is complex
       n = fftOrder;
       real = new double[n];
       imag = new double[n];
       realnxn = new double[n];
       imagnxn = new double[n];
       phaseDer = new double[n];
       realCopy = new double[n];
       imagCopy = new double[n];
       realnxnCopy = new double[n];
       imagnxnCopy = new double[n];
       phase = new double[n];
       bool status = true;

       dOmega = TWO_PI / n;
       phaseUnwrapped.reserve(n * 2); // a reasonable overestimated size.
       standardFreq.reserve(n * 2);

       // I choose using complex fft because the realfft of Mayer seems buggy for the imaginary part (opposite sign) and the value at Pi.
       for(int j = 0; j < n; j++) {
         if (j < nbPoints) real[j] = realCopy[j] = xr[j];
         else real[j] = realCopy[j] = 0.;
         imag[j] = imagCopy[j] = 0.; // xi[j]; if there is an imaginary part
       }

	    for (int j = 0; j < n; j++) {
   	   if (j < nbPoints) realnxnCopy[j] = realnxn[j] = real[j] * j;
         else realnxnCopy[j] = realnxn[j] = 0.;
         imagnxnCopy[j] = imagnxn[j] = 0.; // imag[j] * j; if there is an imaginary part
       }

	    // FFT to compute the spectrum
 	    fft(n, real, imag);
 	    // FFT to compute the phase derivative
 	    fft(n, realnxn, imagnxn);

 	    // Computation of the derivative of arg[X(exp jw)]
 	    for (int j = 0; status && j <= n / 2; j++) {
         double module = real[j] * real[j] + imag[j] * imag[j];
         status  = (fabs(module) > 1e-10);
         if (status) phaseDer[j] = -(real[j] * realnxn[j] + imag[j] * imagnxn[j]) / (real[j] * real[j] + imag[j] * imag[j]);
       }
       if (status) {
              standardFreq.push_back(0.);
              phaseUnwrapped.push_back(0.);
              dOmega = TWO_PI / n;

              for (int j = 1; j <= n / 2; j++)
                   unwrap((j - 1) * dOmega, j * dOmega, phaseUnwrapped.back(), phaseDer[j - 1], phaseDer[j], real[j], imag[j]);

              rearrange();
              removeLinearPhase();
         }
         else for (int j = 1; j <= n / 2; j++) phase[j] = 0.;
}

double RealPhaseUnwrapping::unwrap(
  double a,
  double b,
  double phi_a,
  double phi_der_a,
  double phi_der_b,
  double realXb,
  double imagXb)
{
    double phaseIncrement = (b - a) * (phi_der_a + phi_der_b) / 2.,
           arg1PrincipalValue = atan2(imagXb, realXb),
           argIntegral = phi_a + phaseIncrement,
           deltaArg = argIntegral - arg1PrincipalValue,
           argErrorRounded = floor(deltaArg / TWO_PI + 0.5),
           argErrorRoundedPi = floor (deltaArg / M_PI + 0.5),
           argError = fabs(deltaArg - argErrorRounded * TWO_PI),
           argErrorPi = fabs(deltaArg -  argErrorRoundedPi * M_PI);
    if (argError < ARGDerTHRESHOLD2) {
       double phi = arg1PrincipalValue + argErrorRounded * TWO_PI;
       standardFreq.push_back(b);
       phaseUnwrapped.push_back(phi);
       return(phi);
    }
    else if (argErrorPi  < ARGDerTHRESHOLD2) {
       double phi = arg1PrincipalValue +  argErrorRoundedPi * M_PI;
       standardFreq.push_back(b);
       phaseUnwrapped.push_back(phi);
       return(phi);
    }
    else if (fabs(b - a) < 1e-5) {   // this should not happen...
         cout << "trop de subdivisions ";
      return phi_a;
    }
    else {
      double c = (a + b) /2.,
             XofjwReal, XofjwImag, XofjwRealnxn, XofjwImagnxn,
             phi_der_c, phi_c;

	   realGoertzel(realCopy, &XofjwReal, &XofjwImag, n, c);
//      bruteForceFT(realCopy, imagCopy, &XofjwReal, &XofjwImag, n, c);
      realGoertzel(realnxnCopy, &XofjwRealnxn, &XofjwImagnxn, n, c);
//      bruteForceFT(realnxnCopy, imagnxnCopy, &XofjwRealnxn, &XofjwImagnxn, n, c);
      phi_der_c = -(XofjwReal * XofjwRealnxn + XofjwImag * XofjwImagnxn) / (XofjwReal * XofjwReal + XofjwImag * XofjwImag);
      phi_c = unwrap(a, c, phi_a, phi_der_a, phi_der_c, XofjwReal, XofjwImag);
      return(unwrap(c, b, phi_c, phi_der_c, phi_der_b, realXb, imagXb));
	}
}





