/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

// fichier pmlispit.cpp
//

#include "common.h"
#pragma hdrstop

#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include "snoori.h"

#include "spectro.h"
#include "pitchm.h"
#include "signal.h"
#include "spectre.h"
#include "outils.h"

/*
 * Nombre de transitions sourd-voise et voise-sourd
 *   en remontant un chemin
 */
static int nbTrans( 
	pdynVec *tabLis,	/* tableau des points du lissage */
	int i,	/* date */
	int l,	/* numero du point de programmation dynamique (0, 1 ou 2) */
	int n)	/* profondeur de remontee */
{     
  int j, k,
      nTrans;
  for (j = i, k = l, nTrans = 0; j > max(0, i - n);)
  { int antDat = tabLis[j][k].antDat,
        antTyp = tabLis[j][k].antTyp;
    nTrans += (((antTyp && !k) || (!antTyp && k)) ? 1 : 0);
    j = antDat;
    k = antTyp;
  }
  return(nTrans);
}

/*
 * Distance entre deux valeurs de pitch
 */
static float distF0( 
	float f01, float f02, /* valeurs du pitch */
	int i1, int i2)	/* dates correspondantes */
{
  if (i1 == i2) return(0.);
  else return(((float)(f01 - f02)) / ((float)(i1 - i2)) / DEPPITCH);
}

/*
 * Lissage du pitch par programmation dynamique
 */
void lisPitch( 
	pdynVec  *tabLis,
	int n,	/* nombre de valeurs de pitch */
	struct paramPitch& param)
{
  int i;
  int scanVois = 4;	/* delai pendant lequel on recherche d'eventuelles transitions de voisement */

  if (debug==DEBUG_PM)
    fprintf(ficpm,"lisPitch()\n");

  for (i = 1; i < n; i++)
  { int l;
    for (l = 0; l < 4; l++)
    { float dMin = INFINI_F;
      int j,		/* date de la trame d'ou on vient */
          jMin = -1,	/* coordonnees du meilleur antecedent */
          kMin = -1,
	  dirMin = 0.;	/* direction avec le dernier point raccorde au point etudie */
      // On elimine tout de suite les valeurs de F0
      //   fantaisistes:
      //     F0 plus petit que 0.4 * F0 median,
      //     F0 plus grand que 2.5 * F0 median 
      if (l && !dans(tabLis[i][l].f0
		     , 0.4 * pitchBrutMedian
		     , 2.5 * pitchBrutMedian))
      {	dMin = INFINI_F;
	jMin = -1;
	kMin = -1;
	dirMin = 0.;
      }
      else for (j = max(0, i - param.mem); j < i; j++)
      {	int k;
	for (k = 0; k < 4; k++) if (tabLis[j][k].f0 > -1.)
	{ // changement d'etat sourd-voise ou
	  //   voise-sourd (trait si le nombre
	  //   de transitions d'etats de voisement
	  //   n'est pas trop grand)
	  if ((k && !l) || (!k && l))
	  { if (nbTrans(tabLis 
			, j, k, scanVois) < 1)
	    { float bonuss, d;
	      bonuss = tabLis[i][l].bonus;
	      d = (k) ? distF0(tabLis[j][k].f0, 0., j, i) / 5. : distF0(0., tabLis[i][l].f0, j, i) / 5.;
	      d = fabs(d);
	      if (-bonuss + param.alpha * d + tabLis[j][k].dcum < dMin) {
		dMin = -bonuss + param.alpha * d + tabLis[j][k].dcum ;
		jMin = j;
		kMin = k;
		dirMin = 0.;	// on ne prend pas en
		       // compte la direction lors d'un
		       // changement d'etat 
	      }
	    }
	  }
	  // on est dans le cas de non voisement,
	  //   la distance entre 2 points consecutifs
	  //   est nulle 
	  else if (!k)
	  { float bonuss = tabLis[i][l].bonus;
	    if (-bonuss + tabLis[j][k].dcum  < dMin)
	    { dMin = -bonuss + tabLis[j][k].dcum ;
	      jMin = j;
	      kMin = k;
	      dirMin = 0.;
	    }
	  }
	  // on est maintenant dans le cas de 3 points
	  //   consecutifs voises 
	  else
	  { float d, bonuss;
	    d = distF0(tabLis[j][k].f0, tabLis[i][l].f0, j, i);
	    bonuss = tabLis[i][l].bonus;

	    // on penalise tres fortement les doublements
	    //   ou les divisions par 2 
	    if (fabs( fabs(log((double) tabLis[j][k].f0
				       / tabLis[i][l].f0))
		      / log(2.) - 1.) < 0.15) d = 4 * d;

	    if (-bonuss + param.alpha * fabs(d) 
		+ param.beta * fabs(d - tabLis[j][k].dir)
		+ tabLis[j][k].dcum  < dMin)
	    { dMin = -bonuss + param.alpha * fabs(d) 
		     + param.beta * fabs(d - tabLis[j][k].dir) + tabLis[j][k].dcum ;
	      jMin = j;
	      kMin = k;
	      dirMin = d;
	    }
	  }
	}
      }
      tabLis[i][l].antDat = jMin;
      tabLis[i][l].antTyp = kMin;
      tabLis[i][l].dir = dirMin;
      tabLis[i][l].dcum = dMin;
    }
  }
}

/*
 * Recherche du meilleur parcourt dans le treillis du pitch.
       extrait pitchCorrige de tabLisGlob
 */
void meilPitch( 
	INTarray& pitchCorrige, // tableau du pitch corrige 
    	int n, 		// nombre de valeurs de pitch 
    	int mem)	// memoire du lissage 
{
  int iMin = -1, kMin = -1,	/* date et type du point qui fournit le meilleur resultat */
      i, k;
  float dMin = INFINI_F;

  if (debug==DEBUG_PM)
    fprintf(ficpm,"meilPitch()\n");


  pitchCorrige.newsize(n);

  if (n==0) return;

  for (i = max(0, n - mem); i < n; i++)
  { for (k = 0; k < 4; k++)
      if (tabLisGlob[i][k].dcum < dMin)
      { dMin = tabLisGlob[i][k].dcum;
        iMin = i;
        kMin = k;
      }
  }
  for(i = 0; i < n; i++) pitchCorrige[i] = -1;
  pitchCorrige[iMin] = tabLisGlob[iMin][kMin].f0;
  for (i = tabLisGlob[iMin][kMin].antDat
       , k = tabLisGlob[iMin][kMin].antTyp;
       i >= 0;)
  { int ii;
    if ((i<0)||(i>=pitchCorrige.size()))
      affiche_bug("meilpitch() : bad pitch index");
    pitchCorrige[i] = tabLisGlob[i][k].f0;
    ii = i;
    i = tabLisGlob[ii][k].antDat;
    k = tabLisGlob[ii][k].antTyp;
  }
}

/*
 * Pour complter les valeurs du pitch qui ne sont
 * pas encore connues, parce que on autorise le
 * lissage  sauter des points. On considre ici
 * qu'on ne peut pas sauter plus d'un point  chaque
 * fois.
 */
void completePitch(
	INTarray& pitch,  // tableau des valeurs de pitch 
    	int n)	// nombre de valeurs de pitch
{
  int i;

  if (debug==DEBUG_PM)
    fprintf(ficpm,"completePitch()\n");


  for (i = 0; i < n; i++) if (pitch[i] == -1) {
    if (  ((i==0) && (pitch[i+1]==0))
        ||((i==n-1) && (pitch[i-1]==0))
	||((pitch[i-1]==0) && (pitch[i+1]==0)) )
    { // point est encadre par deux points sourds:
      //     on annule le pitch
      pitch[i] = 0;
    }
    else if ( (i==0) && (pitch[i+1]!=0) )
    { // 1er point: on annule le pitch
      pitch[i] = 0;
      /* on met a jour le tableau de lissage avant rectification des debuts de voisement */
      tabLisGlob[i][0].dcum = 0;
      tabLisGlob[i][1].ignore = VRAI;
      if (debug==DEBUG_PM) fprintf_tabLisGlob(ficpm,i);
    }
    else if ( (pitch[i-1]==0) && (pitch[i+1]!=0) )
    { // aprs un point sourd: on annule le pitch
      pitch[i] = 0;
      /* on met a jour le tableau de lissage avant rectification des debuts de voisement */
      tabLisGlob[i][0].dcum = tabLisGlob[i-1][0].dcum;
      tabLisGlob[i][1].ignore = VRAI;
      if (debug==DEBUG_PM) fprintf_tabLisGlob(ficpm,i);
    }
    else if ( (pitch[i-1]!=0) && (pitch[i+1]==0) )  // ajoute le 28/12/2006
    { // devant un point sourd: on annule le pitch
      pitch[i] = 0;
      /* on met a jour le tableau de lissage avant rectification des debuts de voisement */
      tabLisGlob[i][0].dcum = tabLisGlob[i-1][0].dcum;
      tabLisGlob[i][1].ignore = VRAI;
      if (debug==DEBUG_PM) fprintf_tabLisGlob(ficpm,i);
    }
    else if ( (i==n-1) && (pitch[i-1]>0) )
    { // dernier point: on prend  le pitch du point prcdent
      pitch[i] = pitch[i-1];
    }
    else if ( (pitch[i-1]>0) && (pitch[i+1]>0) )
    { // point normal: lissage local
      pitch[i] = IROUND((pitch[i-1] + pitch[i+1]) / 2.);
      tabLisGlob[i][1].ignore = VRAI;
      if (debug==DEBUG_PM) fprintf_tabLisGlob(ficpm,i);
    }
  }
}

/*
 * Affichage du pitch  
 */

void affMeilPitch (
	INTarray& pitch,	// les valeurs du pitch
   int premEchtAbs,   // premier echantillon de calcul du pitch
	int n)		// nombre de valeurs de pitch
{ int i;
  float  xcote, ycote;


  xcote = wspectro->wsona->DXtoLX(3) - wspectro->wsona->DXtoLX(0);
  ycote = wspectro->wsona->DYtoLY(0) - wspectro->wsona->DYtoLY(3);

  for (i = 0; i < n; i++) if (pitch[i] > 0)
  { float x1,y1;
    x1 = premEchtAbs + i * pm.depEcht + pm.fenPitch/2. - xcote/2.;
    y1 = max(pitch[i] / pm.F0Maxscale() * NB_CANAUX  - ycote/2., 0.);
    snorbox(x1, y1, x1 + xcote, y1 + ycote, wspectro->wsona);
  }
  InvalidateRect(wspectro->wsona->HWindow,NULL,TRUE);
}

/*
 * Identifie le No (entre 1 et 3) du candidat f0 qui a ete retenu
 */
static int numF0( 
	int f0,		/* la valeur de f0 retenue */
	pdynPoint *pt)	/* la trame de programmation
	                   dynamique */
{ int i;
  for (i = 1; i < 4 && f0 != pt[i].f0; i++);
  return(i);
}

void fprintf_tabLoc(FILE *ficpm, int i, pdynVec *tabLoc)
{
  int j;

  fprintf(ficpm,"tabLoc[%d]\n",i);
  for (j=0;j<4;j++)
  { pdynPoint *p=(pdynPoint *)&tabLoc[i][j];
    fprintf(ficpm,"... [%d] %.3g %.3g %d %d %d %.3g %.3g %.3g\n"
		,j
		,p->dcum,p->dir
		,p->antDat,p->antTyp,p->ignore
                ,p->bonus,p->cvois,p->f0);
  }
  fflush(ficpm);
}

/*
 * Rectification des attaques de voisement
 */
void rectifiePitch(
	long premEcht, /* premier echantillon du calcul
	                 de pitch */
	INTarray& pitch,  /* tableau des valeurs de pitch */
    	int n,	      /* nombre de valeurs de pitch */
	struct paramPitch& param)	/* parametres de l'extraction du pitch */
{
  struct paramPitch paramLoc;	/* Les parametres locaux de lissage */
  int i;
  float aBattre;	/* distance a battre si on peut rectifier le debut de voisement */

  if (debug==DEBUG_PM)
    fprintf(ficpm,"rectifiePitch()\n");


  /* Initialisation du calcul de spectre */
  initSpecFft(pm.fenPitch, pm.ordreTF);

  paramLoc = param;
  paramLoc.mem = 1;

  for (i = 0; (i<n);
       i++)
    /* On corrige le pitch si on decouvre un debut de voisement significatif */
    if ( !pitch[i]
	&& (i>4) && (i<n-4)
	&& !pitch[i-1] && !pitch[i-2] && !pitch[i-3] && !pitch[i-4] 
	&& pitch[i+1] && pitch[i+2] && pitch[i+3] && pitch[i+4])
    { pdynVec tabLoc[5];
      int j, 
          pitchEst[3][3];	/* Les estimations du pitch apres correction du deplacement */
      float ener[3][3],		/* Les correlations dans la bande de recherche du voisement */
            coefVois[3][3];	/* Les coefficients de voisement pour les 3 candidats */

      if (debug==DEBUG_PM)
	fprintf(ficpm,"rectifiePitch()  i=%d\n",i);
      /* on initialise tabLoc */
      for (j = 0; j < 5; j++)
      { int k;
	for (k = 0; k < 4; k++)
	{ tabLoc[j][k].dcum = INFINI_F / 10.;
	  tabLoc[j][k].dir = INFINI_F / 10.;
	  tabLoc[j][k].antDat = 0;
	  tabLoc[j][k].antTyp = -1;
	  tabLoc[j][k].ignore = 0;
	  tabLoc[j][k].bonus = -INFINI_F / 10.;
	  tabLoc[j][k].cvois = 0.;
	  tabLoc[j][k].f0 = 0.;
	}
      }

      /* on commence par recuperer la distance cumulee au point de depart */
      tabLoc[0][0].dcum = tabLisGlob[i][0].dcum;
      for (j = 1; j < 4; j++) {
	tabLoc[0][j].dcum = INFINI_F / 10.;
	tabLoc[0][j].f0 = 1.;
      }
      
      /* puis la frequence du point No 3 (c-a-d le premier point non touche par les corrections) 
         Eventuellement si ce point est saute par le lissage  on prend le suivant. */
      if (!tabLisGlob[i+4][1].ignore)
      { int jBon = numF0(pitch[i + 4],(pdynPoint  *)tabLisGlob[i + 4]);
         tabLoc[4][1] = tabLisGlob[i+4][jBon];
      }
      else
        tabLoc[4][1] = tabLisGlob[i+5][numF0(pitch[i + 5],(pdynPoint *)tabLisGlob[i + 5])];
      aBattre = tabLoc[4][1].dcum;

      tabLoc[4][2].f0 = tabLoc[4][3].f0 = -1;
      tabLoc[4][0].bonus = 0.;

      /* On recalcule les 3 trames de pitch incertaines */
      calUnPitch(premEcht + (i + 1. + 5./4.) * pm.depEcht
              , pm.fenPitch
	      , pitchEst[0]
	      , ener[0], coefVois[0]
	      , param);
      calUnPitch(premEcht + (i + 2. + 3./4.) * pm.depEcht
	      , pm.fenPitch
	      , pitchEst[1]
	      , ener[1], coefVois[1]
	      , param);
      calUnPitch(premEcht + (i + 3. + 1./2.) * pm.depEcht
	      , pm.fenPitch
	      , pitchEst[2]
	      , ener[2], coefVois[2]
	      , param);

      /* On calcule les boni qui vont avec les 2 points recalcules */
      for (j = 0; j < 3; j++)
      {	int k;
	for (k = 0; k < 3; k++)
	{ // creation d'un point de programmation
	  //   dynamique voise 
	  tabLoc[j + 1][k + 1].antDat = -1;
	  tabLoc[j + 1][k + 1].f0 = pitchEst[j][k];
	  if (  (tabLoc[j + 1][k + 1].f0>0)
	      &&(!tabLisGlob[i+j+1][1].ignore) ) 
	    tabLoc[j + 1][k + 1].bonus
	      = bonusVoise(ener[j][k], coefVois[j][k]
	                  , enerPeriod[i+1+j]);
	  else tabLoc[j + 1][k + 1].bonus = 0.;
	} 
	// On neutralise le point sourd en lui
	//   donnant un bonus infiniment negatif
	tabLoc[j + 1][0].antDat = -1;
	tabLoc[j + 1][0].bonus = -INFINI_F / 10.;
      }

      if (debug==DEBUG_PM)
      { int i;
	for (i=0;i<5;i++)
	  fprintf_tabLoc(ficpm,i,tabLoc);
      }
      /* On relisse localement */
      lisPitch(tabLoc, 5, paramLoc);

      if (debug==DEBUG_PM)
      { int i;
        for (i=0;i<5;i++)
          fprintf_tabLoc(ficpm,i,tabLoc);
      }

      // Si c'est meilleur que ce qu'on pouvait faire,
      //   on tient compte de la rectification
      if (tabLoc[4][1].dcum < aBattre) {
	   	int k3, k2;
			pitch[i+3] = tabLoc[3][k3 = tabLoc[4][1].antTyp].f0;
			pitch[i+2] = tabLoc[2][k2 = tabLoc[3][k3].antTyp].f0;
         if (k3 == -1 || k2 == -1 || tabLoc[2][k2].antTyp == -1)
                                        affiche_erreur("unconsistent prog dyn in RectifiePitch");
			pitch[i+1] = tabLoc[1][tabLoc[2][k2].antTyp].f0;
			/* Et eventuellement si a i+3 on n'avait qu'une estimation du pitch */
				if (tabLisGlob[i+4][1].ignore)
	 			 pitch[i+4] = (pitch[i+3] + pitch[i+5]) / 2;
      }
    }
  freeSpecFft();
}

