/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

// fichier pmvoisem.cpp
//

#include "common.h"
#pragma hdrstop

#include <math.h>
#include <stdio.h>
#include <malloc.h>

#include "spectro.h"
#include "pitchm.h"
#include "outils.h"
#include "signal.h"

/*
 * Energie dans la bande fbas-fhaut.
 */
void ener(
   spectres &spectroTab,	 // spectra on which the energy is calculated
	FLOATbuf& res,  // tableau de l'energie calculee
   int premEcht,    // premier et dernier echantillon de la region.
   int derEcht,
	float fbas,  // frequences limites en Hz 
	float fhaut) //     de la bande de frequence
{
   int debut, fin,
       cbas = IROUND(freqToEcr(fbas)), chaut = IROUND(freqToEcr(fhaut)),
       j, k;

	debut = premEcht / spectroTab.depl;
   fin = (derEcht + 1 - FENFFT256) / spectroTab.depl;

   if (debug==DEBUG_PM) {
     fprintf(ficpm,"ener() cbas=%d chaut=%d\n",cbas,chaut);
     fprintf(ficpm,"... res.size()=%ld\n",res.size());
   }

  /* calcul de l'energie */
  for (j = debut, k = 0; j <= fin; j++, k++) {
    int c;
    float s;
    for (c = cbas, s = 0.; c <= chaut; c++) {
      float x = spectroTab[j][c];
      s += ((x > SILENCE) ? x - SILENCE : 0.);
    }
    s /= (chaut - cbas + 1);
    res[k] = s;
    if (debug==DEBUG_PM) {
     fprintf(ficpm,"ener() res[%d]=%.3g\n",k,s);
      fflush(ficpm);
    }
   }
}

/*
 * Calcul de l'histogramme du voisement.
 */
void histoVois( 
	FLOATbuf& s,  // le tableau dont on construit l'histogramme 
	int debut,  // limites du sous-tableau
	int fin,    //     a balayer
        INTbuf& histo)
{ int j;
  long n=histo.size();

  for (j = 0; j < n; j++) histo[j] = 0;

  for (j = debut; j <= fin; j++)
  { if (s[j] < 1.) histo[0]++;
    else if (s[j] >= (n-1.)) histo[n-1]++;
    else histo[(int)s[j]]++;
  } 
}

/*
 * Recherche de la valeur au-dessus de laquelle il n'y a que p%
 * des observations non nulles.
 */
int vsupHisto( 
	INTbuf& h, 	// histogramme  
	int effectif,	// nombre d'individus total 
	float p)
{ int i,
      ncum,	/* nombre de valeurs depuis la valeur sup. de l'histogramme */
      nNonNuls = effectif - h[0],
      ip = IROUND(nNonNuls * p); 
  int nClas=(int)h.size(); 

  for (i = nClas - 1, ncum = 0;
       (ncum<ip) && (i>0); // bug corrig (i>=0)
       i--) ncum += h[i];
  return(i);
}
  
static float maxVoise;
static float maxSourd;
static float seuilSourd;
static float bonusSourdMax;
/*
 * Reglage des fonctions de bonus.
 */

extern bool show_warnings;

void reglageBonus(
	float maxVoisement, // valeur maximale observee
	                    // du critere de voisement
	float, //medVoisement, // voisement median
	float maxEnerVois)  // energie maximale dans
	                    // la bande de voisement
{
  /* Ces reglages sont le fruit d'experiences,
     ils peuvent etre affines ou remis en cause. */

  maxVoise = maxVoisement - MAXVOISEMENT;
  maxSourd = maxVoisement - MAXSOURD;
  seuilSourd = max(0, maxEnerVois - 20);
  bonusSourdMax = maxVoisement - 2./ 3. * MAXVOISEMENT;
  if (maxVoise < 0.)
  { if (!noPitchMessage && show_warnings) affiche_warning(ml(7202,"reglageBonus() : Attention je ne peux pas rgler\n\
... correctement le seuil de voisement car le\n\
... signal est beaucoup trop faible, vous devriez\n\
... enregistrer un signal plus intense !!!"));
    maxSourd = 1.;
    maxVoise = 2.;
  }
  else if (maxSourd < 0.)
  { if (!noPitchMessage && show_warnings) affiche_warning(ml(7203,"reglageBonus() : Attention je ne peux pas rgler\n\
... correctement le seuil de voisement car le\n\
... signal est trop faible, vous devriez\n\
... enregistrer un signal plus intense"));
    maxSourd = 2.;
    maxVoise = maxVoise + 2;
  }
  else if (maxSourd < 2.)
  { if (!noPitchMessage && show_warnings) affiche_warning(
ml(7201,"reglageBonus() : Attention j'ai du mal  rgler\n\
... correctement le seuil de voisement car le\n\
... signal est relativement faible"));
    maxVoise = maxVoise + 2.;
    maxSourd = maxSourd  + 2.;
  }

  if (debug==DEBUG_PM)
  { fprintf(ficpm,"reglageBonus()\n");
    fprintf(ficpm,"... maxVoisement=%.3g\n",maxVoisement);
    fprintf(ficpm,"... maxEnerVois=%.3g\n",maxEnerVois);
    fprintf(ficpm,"... maxSourd=%.3g\n",maxSourd);
    fprintf(ficpm,"... maxVoise=%.3g\n",maxVoise);
    fprintf(ficpm,"... seuilSourd=%.3g\n",seuilSourd);
    fprintf(ficpm,"... bonusSourdMax=%.3g\n",bonusSourdMax);
  }
}

/*
 * Contribution du coefficient de voisement au bonus
 */
static float contribVois(
	float cvois, 
     	float x)
{ float res;
  /* On renforce l'energie observee si le coefficient de voisement est tres bon. */
  if (cvois < 10.) res=x * (2. - cvois / 10.);
  else if (cvois <= 40.) res=x; 
  /* On affaiblit l'energie observee si le coefficient de voisement est mauvais. */
  else if (cvois >= 80.) res=0.; 
  else res=(80. - cvois) / 40. * x;
  if (debug==DEBUG_PM)
    fprintf(ficpm,"contribVois(%.3g,%.3g)=%.3g\n"
                 ,cvois,x,res);
  return res;
}

/*
 * Bonus de voisement
 * Le bonus est corrige en fonction du pitch median et est concu
 * par defaut pour un pitch de 100 Hz.
 */
float bonusVoise( 
	float xx,
      	float coefVois,
     	float en) // energie dans la bande de voisement 
{ float res,x;
  if (en <= seuilSourd / 2.)
    x = (seuilSourd > EPSILON) ? xx * en / seuilSourd : 0.;
  else x = contribVois(coefVois, xx);
  if (x <= EPSILON) res=0.;
  if (x > maxVoise / 2.) res=x * pitchBrutMedian / 100.;
  else res=0.5 * (sin(-_PI/2.+ (x/maxVoise)*_PI) + 1.)
               * maxVoise * pitchBrutMedian / 100.;
  if (debug==DEBUG_PM)
    fprintf(ficpm,"bonusVoise(%.3g,%.3g,%.3g)=%.3g\n"
                 ,x,coefVois,en,res);
  return res;
}

/*
 * Bonus de non voisement
 * Le bonus est corrige en fonction du pitch median et est concu
 * par defaut pour un pitch de 100 Hz.
 */
float bonusSourd( 
	float x,
	float coefVois,
      	float en) // energie dans la bande de voisement 
{ float res;

  if (en <= seuilSourd) x = (seuilSourd > EPSILON) ? x * en / seuilSourd : 0.;
  else x = contribVois(coefVois, x);
  if (x <= EPSILON) return(bonusSourdMax * pitchBrutMedian / 100.);
  if (x > maxSourd) return(0.);
  res=cos(x / maxSourd * _PI / 2.)
       * bonusSourdMax * pitchBrutMedian / 100.;
  if (debug==DEBUG_PM)
    fprintf(ficpm,"bonusSourd(%.3g,%.3g,%.3g)=%.3g\n"
                 ,x,coefVois,en,res);
  return res;
}


//
// Rearrangement de la correlation pour projeter la courbe de
// correlation autour de f0
//      
static void rearrangeCorrel( 
	FLOATbuf& res_,
	FLOATbuf& c_,
        int f0,
	int ninf, int nsup,
	float seuil)
{ int k;
  float *res=(float *)&res_[0];
  float *c=(float *)&c_[0];

  for (k = 0; k <= f0; k++) res[k] = 0.;
  for (k = 0; k < nsup / f0; k++)
  { int i;
    for (i = max(ninf, k * f0);
	 i < min(nsup, (k + 1) * f0);
	 i++)
    { float x = (c[i] > seuil) ? c[i] - seuil : 0.;
      if ((i % f0) < f0 / 2)
        res[f0/2 + i % f0 + (f0 % 2)] += x;
      else res[i % f0 - f0/2] += x;
    }
  }
}
//
// Calcul du moment d'ordre 2 de la correlation.
//
static float moment2Correl(
	FLOATbuf& t_,  // la distribution a etudier
	int xi,	    // point par rapport auquel
	            // on calcule le moment 
	int imin,  // le domaine de
	int isup)  //     s a etudier 
{ int i;
  float s2, s0; 
  float *t=(float *)&t_[0];

  for (i = imin, s0 = 0., s2 = 0.; i <= isup; i++)
  { float x = t[i];
    s0 += x;
    s2 += x * SQR(i - xi);
  }
  return(sqrt((s0 > EPSILON) ? s2 / s0 : 0.));
}
/*
 * Calcul du coefficient de voisement sur la courbe de correlation spectre-peigne.
 * Le coefficient est d'autant plus faible que le pic representant F0 est un bon
 * pic.
 */
float coefVoisCorrel( 
	FLOATbuf& correl,
	int f0)
{ FLOATbuf correlBis_(NBPTPREL,"correlBis in coefVoisCorrel()");
  float *correlBis=(float *)&correlBis_[0];  
  float vmax = correl[f0] 
       , somCorrel,
	m1, m2, m;
  int i;

  rearrangeCorrel(correlBis_, correl, f0, 50
               , f0 + max(200, f0) + 5, vmax * 0.6);

  for (i = 0, somCorrel = 0.; i < f0; i++)
    somCorrel += correlBis[i];

  for (i = (int)(f0 * 0.15), m1 = 0.; i < (int)(f0 * 0.3); i++) m1 += correlBis[i];
  for (i = (int)(f0 * 0.7), m2 = 0.; i < (int)(f0 * 0.85); i++) m2 += correlBis[i];
  for (i = (int)(f0 * 0.4), m = 0.; i < (int)(f0 * 0.6); i++) m += correlBis[i];

  m1 = max_f(correlBis_, (int)(f0 * 0.15), (int)(f0 * 0.3));
  m2 = max_f(correlBis_, (int)(f0 * 0.7), (int)(f0 * 0.85));
  m = ((m1 + m2) < EPSILON) ? m = 15. : m / (m1 + m2);
  /*
  return(min(INFINI_F/1000
             , moment2Correl(correlBis_, f0 / 2
		     , (int)(f0 * 0.15),  (int)(f0 * 0.85))
               / (correlBis[f0/2]+EPSILON)
               * somCorrel
	       / m
	    )
        );
  */
  return(  min(INFINI_F/1000
               ,moment2Correl(correlBis_, f0 / 2
		     , (int)(f0 * 0.15)
		     , (int)(f0 * 0.85) )
              )
	 / max(EPSILON*1000
	       ,(correlBis[f0/2]+EPSILON)
              )
	 * min(INFINI_F/1000
	       ,somCorrel
              )
	 / max(EPSILON*1000
	       ,m
	      )
	);
}

