/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

// fichier regrespa.h
//

#ifndef _REGRESPA_H
#define _REGRESPA_H


/*
 * Prototype de regression parabolique. 
 * Ce fichier prototype permet de construire des fonctions de regression parabolique
 * avec les fonctions d'initialisation.
 * Ces fonctions sont destinees a etre utilisees de nombreuses fois, c'est pour
 * cela que les coefficients auxiliaires sont initialises et sont propres a une
 * fonction de regression. 
 * Si le nombre de points est variable il est enormement preferable
 * d'utiliser les fonctions dans regrespar.c
 * UTILISATION:
 * #define N        --> nombre de points de la regression  		ex: 12
 * #define REGRES   --> nom de la fonction de regression   		ex: regres12parab
 * #define INIT_RESGRES_PARAB --> nom de la fonction d'initialisation	ex: initRegres12
 */
#define EPSILON	1.e-6

static float s1, s2, s3, s4, 			/* somme des x, des x^2, des x^3, des x^4 */
             a11, a12, a13, a22, a23, a33,	/* coeff de la matrice symetrique de minimisation */
	     det;				/* determinant de la matrice precedente */

/*
 * Regression parabolique sur n points, les n points etant regulierement
 * espaces de 1 en abscisse et commencant a 0.
 */
void REGRES(float *a, float *b
	   ,float *c, float *y)
	//float *a,*b,*c, // coefficients de la parabole
	//                       ax^2 + bx + c 
	//  y[];	  // tableau des abscisses
	//                       des n points 
{
  float s01, s11, s21;	/* somme des y, des xy, des x^2y */
  int i;

  for (i = 0, s01 = 0., s11 = 0., s21 = 0.; i < N; i++) {
    s01 += y[i];
    s11 += i * y[i];
    s21 += i * i * y[i];
  }

  *a = (s21 * a11 - s11 * a12 + s01 * a13) / det;
  *b = (-s21 * a12 + s11 * a22 - s01 * a23) / det;
  *c = (s21 * a13 - s11 * a23 + s01 * a33) / det;
}

/*
 * Initialisation de sommes et de produits apparaissant dans le calcul
 * d'une regression parabolique.
 */
void INIT_RESGRES_PARAB()
{
  int i;

  s1 = s2 = s3 = s4 = 0.;
  for (i = 0; i < N; i++) {
    s1 += i;
    s2 += i * i;
    s3 += i * i * i;
    s4 += i * i * i * i;
  }

  a11 = s2 * N - s1 * s1;
  a12 = s3 * N - s1 * s2;
  a13 = s3 * s1 - s2 * s2;
  a22 = s4 * N - s2 * s2;
  a23 = s4 * s1 - s2 * s3;
  a33 = s4 * s2 - s3 * s3;

  det = s4 * a11 - s3 * a12 + s2 * a13;
}

#endif
