/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

#include <math.h>

#define EPSILON 1e-6
#define max(A,B) (((A)>(B)) ? (A) : (B))
#define min(A,B) (((A)>(B)) ? (B) : (A))
#define max3(A,B,C) (((A) < (B)) ? max(B,C) : max(A,C))
#define min3(A,B,C) (((A) < (B)) ? min(A,C) : min(B,C))

/*
-- RGB, each 0 to 255
-- H = 0.0 to 360.0 (corresponding to 0..360.0 degrees around hexcone)
-- S = 0.0 (shade of gray) to 1.0 (pure color)
-- V = 0.0 (black) to 1.0 (white)

-- Based on C Code in "Computer Graphics -- Principles and Practice,"
-- Foley et al, 1996, pp. 592,593.
*/


void rgb2hsv(
   int r,
   int g,
   int b,
   float *h_,
   float *s_,
   float *v_)
{
  float minVal = min3(r, g, b),
        v = max3(r, g, b),
        Delta = v - minVal,
        s, h;

  // Calculate saturation: saturation is 0 if r, g and b are all 0
  s = (fabs(v) < EPSILON) ? 0.0 : Delta / v;

  if (fabs(s) < EPSILON)
    h = 0.0;     // Achromatic: When s = 0, h is undefined but who cares
  else if (fabs(r - v) < EPSILON)      	// Chromatic
    h = 60.0 * (g - b) / Delta;      			// between yellow and magenta [degrees]
  else if (fabs(g - v) < EPSILON)  			// between cyan and yellow
    h = 120.0 + 60.0 * (b - r) / Delta;
  else h = 240.0 + 60.0 * (r - g) / Delta; // between magenta and cyan
  if (h < 0.0) h = h + 360.0;
  // return a list of values as an rgb object would not be sensible
  *h_ = h;
  *s_ = s;
  *v_ = v / 255.;
}

void hsv2rgb(
   float h,
   float s,
   float v,
   int *r_,
   int *g_,
   int *b_)
{
  float f, p, q ,t, r, g, b, hTemp;
  int i;

  if (fabs(s) < EPSILON) { //  color is on black-and-white center line
    r = v;           //  achromatic: shades of gray
    g = v;           //  supposedly invalid for h=0 but who cares
    b = v;
  }
  else if (fabs(h - 360.0) < EPSILON) // chromatic color
    hTemp = 0.0;    //  360 degrees same as 0 degrees
  else hTemp = h;

  hTemp = hTemp / 60.0; 	// h is now in [0,6)
  i = (int) hTemp;  			// largest integer <= h
  f = hTemp - i;        	// fractional part of h

  p = v * (1.0 - s);
  q = v * (1.0 - (s * f));
  t = v * (1.0 - (s * (1.0 - f)));

  switch (i) {
  case 0:
        r = v;
        g = t;
        b = p;
     break;
  case 1:
        r = q;
        g = v;
        b = p;
     break;
  case 2:
        r = p;
        g = v;
        b = t;
     break;
  case 3:
        r = p;
        g = q;
        b = v;
     break;
  case 4:
        r = t;
        g = p;
        b = v;
     break;
   case 5:
        r = v;
        g = p;
        b = q;
     break;
   }
  *r_ = r * 255;
  *g_ = g * 255;
  *b_ = b * 255;
}