/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

// fichier spline.cpp
//

#include "common.h"
#pragma hdrstop

#include <math.h>
#include <malloc.h>

#include "polypoint.h"
#include "guiklat.h"

/* pour un reel tl de [a,b], calcule
   l'indice i tel que tl in [u[i],u[i+1] et calcule
   les valeurs des fonctions
   de base des splines non nulles en tl
   la valeur retournee est l'indice i
   */
static int bspline( 
     double tl,
     double bs[],
     double a, double b,
     int n, int degre)
{
    double h,r,p,q;
    int i;

    /* calcul du pas */
    h= (b-a)/n;
    r = (tl-a)/h;
    i = (int)r;

    if (fabs(tl-b) < 1.e-10 || i==n ) i=n-1;
    /* Calcul des coordonnees relatives dans [u[i],u[i+1]]*/
    p=r-i;
    q=1-p;

    /* determination des valeurs des fonctions spline non nulles en tl */

    switch (degre) {
    case 1: {
	bs[1]=q;
	bs[0]=p;
      }break;
    case 2: {
	bs[2]=q*q/2.;
	bs[1]=(2*p*q+1)/2.;
	bs[0]=p*p/2.;
      }break;
    case 3: {
	bs[3]=q*q*q/6.;
	bs[2]=(3*q*(p*q+1)+1)/6.;
	bs[1]=(3*p*(p*q+1)+1)/6.;
	bs[0]=p*p*p/6.;
      }break;
    }
    return i;
}


 

/* resolution de AX=B pour une matrice A SYMETRIQUE CARREE
   DEFINIE POSITIVE de taille n
   La matrice A n'est PAS CONSERVEE a l'issue de la procedure */

static void choleski( 
     double **A,
     double B[],
     int n,
     double X[])
{ int i,j,k;
  double s;

if (debug==3)
{ Wmsg->wr("choleski");
  char ch[80];
  sprintf(ch,"... n=%d",n);
  Wmsg->wr(ch);
}
  /* etape de factorisation */
  for (j=0;j<n;j++)
    for (i=j;i<n;i++)
    { for(k=0;k<j;k++)
	A[i][j] = A[i][j]-A[i][k]*A[j][k];
      if (i==j)
      { if (A[i][i]>0) A[i][i]= sqrt(A[i][i]);
	else
	{ affiche_bug(
	    "choleski(): matrice non definie positive");
	  return;
	}
      }
      else
	A[i][j]=A[i][j]/A[j][j];
    }

  for (j=0;j<n;j++)
  { s=0.;
    for(k=0;k<j;k++) s += A[j][k]*X[k];
    X[j]= (B[j]-s)/A[j][j];
  }

  for(j=n-1;j>=0;j--)
  { s=0.;
    for (k=j+1;k<n;k++) s +=A[k][j]*X[k];
    X[j]=(X[j]-s)/A[j][j];
  }
}
  

int bsreg(int nd, 	// nombre de points donnes 
	  double T[],	// tableau des abscisses des donnees 
	  double Z[],	// tableau des ordonnees des donnees 
	  int degre, // degre du polynome d'interpolation 
	  double a, double b, // Intervalle de travail
                            // temporel en chantillons 
	  int n,	// nombre de sous intervalles de [a,b]
	  double tau,	// parametre de regularisation 
	  int ns,	// nombre de points ou est
	                //     calculee la fct spline 
	  double alfa[], // coefficient sur la base de spines
	  double X[],	// coordonnees de la spline
	  double Y[])	//     resultat discretisee 
{   int i,j,j0,k0,j1,k1,l;
    double hs,tl,som;
    double **M;
    double *beta;
    double c[4];
    double bs[4];/* comme l'indexation par un nb <0 n'est pas possible
		    dans un tableau bs[0] correspondra a bs[i]
		    bs[1] correspondra a bs[i-1]
		    bs[2] correspondra a bs[i-2]
		    bs[3] correspondra a bs[i-3]
		    correspond au calcul en un point donnee des valeurs des
		    fonctions de la base ne s'annulant pas
		    (4 pour des splines d'ordre 3,
		    3                          2
		    2                          1)*/

if (debug==3)
{ Wmsg->wr("bsreg");
  char ch[80];
  sprintf(ch,"... nd=%d degre=%d",nd,degre);
  Wmsg->wr(ch);
  sprintf(ch,"... a=%5g b=%5g n=%d",(float)a,(float)b,n);
  Wmsg->wr(ch);
  sprintf(ch,"... tau=%f ns=%d",tau,ns);
  Wmsg->wr(ch);
}
    int status;  /* indicateur d'incident */

    status=0;
    beta = (double *) calloc((unsigned int) (n + degre), (unsigned int) sizeof(double));
    M = (double **) calloc((unsigned int) (n + degre +1), (unsigned int) sizeof(double *));
    for(j=0;j<n+degre+1;j++) {
      M[j] = (double *) calloc((unsigned int) (n + degre+1), (unsigned int) sizeof(double));
      if (!M[j]) {
	  int k;
          for (k = 0; k < j; k++) free(M[k]);
          free(beta); free(M);
	  return(1);
      }
    }

    switch (degre) {
    case 1: {c[0]=1.;c[1]= -1.;};break;
    case 2: {c[0]=1.;c[1]=-2.;c[2]=1.;}break;
    case 3: {c[0]=1.;c[1]= -3.;c[2]=3.;c[3]=-1.;}break;
    }
 

   /* Formation de la matrice M et du second menmbre beta */

    /* partie moindre carres */
    for(l=0;l<nd;l++) {
	if (T[l]<a || T[l]>b) status =1;
	i=bspline(T[l],bs,a,b,n,degre);
	for(j0=degre;j0>=0;j0--) {
	    j1=i-j0+degre;
	    beta[j1] += Z[l]*bs[j0];
	    for(k0=degre;k0>=j0;k0--) { /* j0 a la place de o*/
		k1=i-k0+degre;
		M[j1][k1] +=bs[j0]*bs[k0];
	    }
	}
    }



    for(i=0;i<n+degre;i++)
	if (M[i][i]==0) status=1;

    /* partie regularisation */
    for(i=0;i<n;i++)
	for(j0=0;j0<=degre;j0++)
	    for(k0=0;k0<=j0;k0++)
		M[i+j0][i+k0] += tau*c[j0]*c[k0];

    /* resolution du systeme lineaire */
    choleski(M,beta,n+degre,alfa);

    /* evaluation de la spline */
    hs= (b-a)/(ns-1);
    for(l=0;l<ns;l++) {
	   tl= min(a+l*hs, b);
	   i=bspline(tl,bs,a,b,n,degre);
	   som=0.;
	for(j0=0;j0<=degre;j0++)
	    som += alfa[i+j0]*bs[degre-j0];
	X[l]=tl;
	Y[l]=som;
    }
    free(beta);    
    for(j=0;j<n+degre+1;j++) free(M[j]);
    free(M);

    return status;
}

