/* This file is part of WinSnoori.

 WinSnoori is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 3 of the License.

WinSnoori is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with WinSnoori.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#include "fft_mayer.h"

#define SQR(X) ((X) * (X))

void cepstralSpectrum(
  short *speech, double *smoothedSpectrum, int nbSamples, int fftOrder, double *hammingWindow, double *lifter, int lifterLength, bool preemphasis)
{
   double *real = new double[fftOrder],
          *spectrum = new double[fftOrder],
          scale;

    if (preemphasis) for (int k = 0; k < nbSamples; k++)
		  real[k] =  (speech[k + 1] - speech[k]) * hammingWindow[k];
    else for (int j = 0; j < nbSamples; j++) real[j] =  speech[j] * hammingWindow[j];

    for (int k = nbSamples; k < fftOrder; k++) real[k] = 0.;

    // First FFT to compute the spectrum
    realfft(fftOrder, real);

    // magnitude spectrum calculation
    spectrum[0] = 10. * log10(SQR(real[0]) + 1.e-06);
    spectrum[fftOrder / 2] = 10. * log10(SQR(real[fftOrder / 2]) + 1.e-06);
    for (int k = 1; k < fftOrder / 2; k++)
      spectrum[k] = 10. * log10(SQR(real[k]) + SQR(real[fftOrder - k]) + 1e-06);
    for (int k =  fftOrder / 2 + 1; k < fftOrder; k++)        // copy the first half onto the second half
      spectrum[k] = spectrum[fftOrder - k];

    //  Second FFT to compute the cepstrum
    realifft(fftOrder, spectrum);
    scale = 1. / fftOrder;
    for (int i = 0; i < fftOrder; i++) spectrum[i] *= scale;

    // Liftering and completion to 0
    for (int i = 1; i < lifterLength; i++) {
      spectrum[i] *= lifter[i];
      spectrum[fftOrder - i] *= lifter[i];
    }
    for (int i = lifterLength; i < fftOrder / 2; i++) spectrum[i] = spectrum[fftOrder - i] = 0.;
    spectrum[fftOrder / 2] = 0.;

    // Last FFT to get the smoothed cepstrally smoothed spectrum
    realfft(fftOrder, spectrum);

    // Copy the cepstrally smoothed spectrum
    for (int i = 0; i < fftOrder / 2; i++) smoothedSpectrum[i] = spectrum[i];
    delete []spectrum;
    delete []real;
}

main(int argc, char *argv[])
{
  int windowLength, fftOrder, lifterLength, ninf, nsup;
  float samplingRate;
  FILE *speechFile, *spectraFile;
  if (argc != 8) {
    printf("testCepstre <rawSpeechFile> <SamplingRate (16000)> <windowMs(32)> <fftOrder (512)> <ninfLifterMs(1.25)> <nsupLifterMs(2.5)> <spectraFile>");
    exit(0);
  }

  speechFile = fopen(argv[1], "rb");
  if (!speechFile) {
     printf("File %s cannot be accessed\n", argv[1]);
     exit(0);
  }
  samplingRate = atof(argv[2]);
  windowLength = (int) (atoi(argv[3]) * samplingRate / 1000. + 0.5);
  fftOrder = atoi(argv[4]);
  lifterLength = fftOrder / 2;
  ninf = (int) (atof(argv[5]) * samplingRate / 1000. + 0.5);
  nsup = (int) (atof(argv[6]) * samplingRate / 1000. + 0.5);
  spectraFile = fopen(argv[7], "w");
  if (!spectraFile) {
     printf("File %s cannot be accessed\n", argv[7]);
     exit(0);
  }

  double *hw = new double[windowLength];
  short *speech = new short[windowLength + 1]; // + 1 because of the subtraction of the preemphasis
  double *smoothedSpectrum = new double[fftOrder / 2];
  double *lifter = new double[lifterLength];
  int check = windowLength + 1;

  // Initialization of the lifter
  for (int i = 0; i < ninf && i < lifterLength; i++) lifter[i] = 1.;
  for (int i = ninf; i < nsup && i < lifterLength; i++)
    lifter[i] = 0.5 * (1. + cos (M_PI * (i - ninf + 1) / (nsup - ninf)));
  for (int i = nsup; i < lifterLength; i++) lifter[i] = 0.;

  // Initialization of the Hamming window
  for (int k = 0; k < windowLength; k++)
    hw[k] =(float)(0.54 - 0.46 * cos(M_PI * 2. * k / windowLength));

  // It is JUST FOR FUN, there is no overlapping between two consecutive windows.
  // There should be overlapping for ASR and spectrogram calculation as well.
  while (check == windowLength + 1) {
    fseek(speechFile, sizeof(short) * i * decal, SEEK_SET);
    check = fread(speech, sizeof(short), windowLength + 1, speechFile);
    if (check == windowLength + 1) {
      cepstralSpectrum(speech, smoothedSpectrum, windowLength, fftOrder, hw, lifter, lifterLength, true);
      for (int i = 0; i < fftOrder /2; i++) fprintf(spectraFile,"%.2f ", smoothedSpectrum[i]);
      fprintf(spectraFile,"\n");
    }
  }
  delete [] hw;
  delete [] lifter;
  delete [] speech;
  delete [] smoothedSpectrum;

  fclose(speechFile);
  fclose(spectraFile);
}