// exemple d'utilisation:
// klatt -f 5 -r 1 -i example1.par -o example1.raw -s 16000
#include <iostream>
#include <fstream>

#include <malloc.h>
#include <string.h>
#include "parwave.h"
#include "getopt.h"


#ifndef BORLAND
using namespace std;
#endif

/*
title:  KLATT.C
author: Jon Iles (j.p.iles@cs.bham.ac.uk)
date: 20/4/94
version: 3.03
notes:
modified: 28/7/97 Andr Doherty (Andre.Doherty@loria.fr)

This file contains C code to provide a simple interface to
Dennis Klatt's parwave function. This function, as the name
suggests, converts a frame of parameters into a waveform.
The parwave function itself is a translation of the Dennis's
original Fortran code, as published in JASA. The code was
posted to comp.speech in 1993, and has been updated by me
to fix a number of bugs that had been introduced. Subsequently
major rewrites have been done by Nick Ing-Simmons to improve
efficiency and change the code into a more acceptable ANSI C style,
and by me to remove all global variable references, fix bugs and ensure
portability.

See the README file and man page for more details.

(c) 1993 Jon Iles and Nick Ing-Simmons

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



/* for default sampled glottal excitation waveform */

#define NUMBER_OF_SAMPLES 100
#define SAMPLE_FACTOR 0.00001

/* function prototypes */

//static void usage PROTO((void));
static void usage();

/*
function USAGE

Displays the command line options.
*/

static void usage(){
  cerr << "Options...\n";
  cerr << "-h Displays this message\n";
  cerr << "-i <infile> sets input filename\n";
  cerr << "-o <outfile> sets output filename\n";
  cerr << " If output filename not specified, stdout is used\n";
  cerr << "-q quiet - print no messages\n";
  cerr << "-t <n> select output waveform\n";
  cerr << "-c select cascade-parallel configuration\n";
  cerr << "   Parallel configuration is default\n";
  cerr << "-n <number> Number of formants in cascade branch.\n";
  cerr << "   Default is 5\n";
  cerr << "-s <n> set sample rate\n";
  cerr << "-f <n> set number of milliseconds per frame, default 10\n";
  cerr << "-v <n> Specifies voicing source.\n";
  cerr << "   1=impulse train, 2=natural simulation, 3=sampled natural\n";
  cerr << "   Default is a simulation of natural voicing\n";
  cerr << "-V <filename> Input file of samples for natural voicing.\n";
  cerr << "-F <percent> percentage of f0 flutter\n";
  cerr << "    Default is 0\n";
  cerr << "-r <type> output 16 bit signed integers rather than ASCII\n";
  cerr << "   integers. Type = 1 gives high byte first, Type = 2 gives\n";
  cerr << "   low byte first.\n";
}


int main(int argc,char **argv){
  extern char *optarg;
  //  char c;
  char infile[80];
  char outfile[80];
  ifstream infp;
  ofstream outfp;
//  int result;
  flag done_flag;
  long value;
  int *iwave;
  int isam;
  int icount;
  int par_count;
  int nmspf_def;
  klatt_global_ptr globals;
  klatt_frame_ptr frame;
  long *frame_ptr;
  unsigned char high_byte;
  unsigned char low_byte;
  flag raw_flag;
  flag raw_type;
  char samples_file[80];
  int Npar;
  int loop;

  static int natural_samples[NUMBER_OF_SAMPLES]=
  {
    -310,-400,530,356,224,89,23,-10,-58,-16,461,599,536,701,770,
    605,497,461,560,404,110,224,131,104,-97,155,278,-154,-1165,
    -598,737,125,-592,41,11,-247,-10,65,92,80,-304,71,167,-1,122,
    233,161,-43,278,479,485,407,266,650,134,80,236,68,260,269,179,
    53,140,275,293,296,104,257,152,311,182,263,245,125,314,140,44,
    203,230,-235,-286,23,107,92,-91,38,464,443,176,98,-784,-2449,
    -1891,-1045,-1600,-1462,-1384,-1261,-949,-730
  };


  if(argc==1){
    usage();
    exit(1);
  }

  /* set up default values */

  strcpy(infile,"");
  strcpy(outfile,"");
  strcpy(samples_file,"");

  iwave = (int*) malloc(sizeof(int)*MAX_SAM);
  if(iwave==NULL){
    cerr << "malloc failed\n";
    exit(1);
  }

  globals = (klatt_global_ptr)malloc(sizeof(klatt_global_t));
  if(globals==NULL){
    cerr << "malloc failed\n";
    exit(1);
  }

  frame = (klatt_frame_ptr)malloc(sizeof(klatt_frame_t));
  if(frame==NULL){
    cerr << "malloc failed\n";
    exit(1);
  }

  globals->quiet_flag = FALSE;
  globals->synthesis_model = ALL_PARALLEL;
  globals->samrate = 10000;
  globals->glsource = NATURAL;
  globals->natural_samples = natural_samples;
  globals->num_samples = NUMBER_OF_SAMPLES;
  globals->sample_factor = (float) SAMPLE_FACTOR;
  nmspf_def = 10;
  globals->nfcascade = 0;
  globals->outsl = 0;
  globals->f0_flutter = 0;
  raw_flag = FALSE;

  GetOpt getopt(argc, argv, "i:o:qt:cs:f:v:V:hn:F:r:");
  int option_char;

  while ((option_char = getopt()) != EOF){
    switch(option_char){
    case 'i':
      strcpy(infile,getopt.optarg);
      break;
    case 'o':
      strcpy(outfile,getopt.optarg);
      break;
    case 'q':
      globals->quiet_flag = TRUE;
      break;
    case 't':
      globals->outsl = (flag) atoi(getopt.optarg);
      break;
    case 'c':
      globals->synthesis_model = CASCADE_PARALLEL;
      globals->nfcascade = 5;
      break;
    case 's':
      globals->samrate = atoi(getopt.optarg);
      break;
    case 'f':
      nmspf_def = atoi(getopt.optarg);
      break;
    case 'v':
      globals->glsource = (flag) atoi(getopt.optarg);
      break;
    case 'V':
      strcpy(samples_file,getopt.optarg);
      break;
    case 'h':
      usage();
      exit(1);
      break;
    case 'n':
      globals->nfcascade = atoi(getopt.optarg);
      break;
    case 'F':
      globals->f0_flutter = atoi(getopt.optarg);
      break;
    case 'r':
      raw_flag = TRUE;
      raw_type = (flag) atoi(getopt.optarg);
      break;
    }
  }

  globals->nspfr = (globals->samrate * nmspf_def) / 1000;

  if(strcmp(samples_file,"")!=0){
    ifstream INPUT(samples_file, ios::in);
    INPUT.setf(ios::skipws);

    if(!INPUT){
      cerr << "Can't open voicing samples file\n";
      exit(1);
    }

    //    fscanf(infp,"%i",&(globals->num_samples));
    INPUT >> (globals->num_samples);
    //    fscanf(infp,"%f",&(globals->sample_factor));
    INPUT >> (globals->sample_factor);


    globals->natural_samples = (int*) malloc(sizeof(int)*globals->num_samples);
    if(globals->natural_samples==NULL){
      cerr << "malloc failed for natural samples\n";
      exit(1);
    }

    for(loop=0;loop<globals->num_samples;loop++){
      //      fscanf(infp,"%i",&(globals->natural_samples[loop]));
      INPUT >> (globals->natural_samples[loop]);
    }

    //    fclose(infp);
    INPUT.close();
  }


  if(strcmp(infile,"")==0){
    //    infp = cin;
    // stdin a pour numero de descripteur 0
    cerr << "can't attach standard input anymore\n";
    exit(1);
    // infp.attach(0);
  }
  else{
    char magicNumber[5];
    //infp = fopen(infile,"r");
    infp.open(infile,ios::in);
    if(!infp){
      cerr << "can't open input parameter file\n";
      exit(1);
    }
    infp >> magicNumber;
    if (strcmp(magicNumber, ".kla") && strcmp(magicNumber, ".klf")
        && strcmp(magicNumber, ".kle")) {
      cerr << "this is not a valid Klatt file\n";
      infp.close();
      exit(1);
    }
    else if (!strcmp(magicNumber, ".kla")) Npar = NPAR;
    else if (!strcmp(magicNumber, ".kle")) Npar = NPARLE;
    else Npar = NPARLF;
  }

//  ifstream infp(infile,ios::in);

  if(strcmp(outfile,"")==0){
    //outfp = stdout;
    // stdout a pour numero de descripteur 1
    // outfp.attach(1);
    // globals->quiet_flag = TRUE;
    cerr << "can't attach input file anymore\n";
    exit(1);
  }
  else{
    //outfp = fopen(outfile,"w");
    outfp.open(outfile,ios::out|ios::binary|ios::trunc);

    if(!outfp){
      cerr << "can't open output file\n";
      exit(1);
    }
  }

  icount=0;
  done_flag = FALSE;
  parwave_init(globals, Npar);
  frame_ptr = (long*) frame;

  while(done_flag == FALSE){
    for (par_count = 0; par_count < Npar; ++par_count){
      value = 0;
      infp >> value;
      //      result = fscanf(infp,"%i",(int*)&value);
      frame_ptr[par_count] = value;
    }

	 if (infp.eof())
      done_flag = TRUE;
    else{
      parwave(globals,frame,iwave);

      if(globals->quiet_flag == FALSE){
         cout << "\rFrame" << icount;
         cout.flush();
      }

      for (isam = 0; isam < globals->nspfr; ++ isam){
	     if(raw_flag == TRUE){
	       low_byte = iwave[isam] & 0xff;
	       high_byte = iwave[isam] >> 8;

	       if(raw_type==1){
          	outfp << high_byte << low_byte;
	       // fprintf(outfp,"%c%c",high_byte,low_byte);
          }
	       else{
          	outfp << low_byte << high_byte;
	       // fprintf(outfp,"%c%c",low_byte,high_byte);
	       }
	     }
	     else{
        		outfp << iwave[isam] << "\n";
	     //fprintf(outfp,"%i\n",iwave[isam]);
	     }
      }
      icount++;
    }
  }

    if(strcmp(infile,"")!=0)
      infp.close();


    if(strcmp(outfile,"")!=0)
      outfp.close();


  if(globals->quiet_flag == FALSE)
    cout << "\nDone\n";
}
